#!/usr/local/bin/perl -w
#############################################################################
#
#  Program:   local_normal.pl
#  Function:  To normalise micro-array data using the method of
#             R. Lyne et al. (2003), BMC Genomics 2003, 4:27.
#             *Heavily* borrows code from the normalisation scripts of
#             Juan Mata and Rachel Lyne.
#
#  Author:    Chris Penkett
#  Date:      6 May 2003
#  Address:   The Wellcome Trust Sanger Institute
#             Wellcome Trust Genome Campus
#             Hinxton
#             Cambridge
#             CB10 1SA
#  Telephone: +44 (0)1223 494919
#  Email:     cjp@sanger.ac.uk
#
#############################################################################
#
#  Description:
#
#:  This script normalises micro-array data using the method described in
#:  the following paper:
#:    Rachel Lyne, Gavin Burns, Juan Mata, Chris J Penkett, Gabriella
#:    Rustici, Dongrong Chen, Cordelia Langford, David Vetrie, Jurg Bahler;
#:    BMC Genomics 2003, 4:27 (10 July 2003).
#:    Whole-genome microarrays of fission yeast: characteristics, accuracy,
#:    reproducibility, and processing of array data.
#:
#:  If you use this script, please cite the above reference.
#:
#:  *Heavily* borrows code from the normalisation scripts of
#:  Juan Mata and Rachel Lyne.
#:
#:  This perl script uses GenePix results files as input and discards data
#:  from spots with failed or faulty PCR products by masking them absent.
#:  Data from spots with low array element concentration (as judged by PCR
#:  product staining on gel) or PCR products where the reverse primer is
#:  located 2500-3500 bp from the gene end are flagged marginal.
#:  In the GenePix results file, the 'ID' column is used for the gene
#:  ID and the 'Name' column is used for information about the type or
#:  quality of the PCR spot on the micro-array slide.  In the 'Name' column,
#:  failed or faulty spots are set to 'A' (absent), and spots with low probe
#:  concentration or where the reverse primer is located 2500-3000 bp from
#:  the 3' end of the gene are set to 'M' (marginal).  Also, in the 'Name'
#:  column, 'Q' indicates bacterial controls, 'T' indicates tags or markers,
#:  and 'Y' indicates genes that are controls for cross-hybridisation
#:  (budding yeast genes of different levels of homology in our case).
#:
#:  This information is then used together with the 'Flags' column in the
#:  results file regarding the spot finding algorithm.  For example, spots
#:  not found are given a flag of '-50', and spots masked as bad by the user
#:  are given a flag of '-100' by GenePix.  These flags are masked as being
#:  absent ('A').  For each spot, a combined flag is produced such that any
#:  spot that is absent in either the 'Name' column or the 'Flags' column is
#:  given an overall flag of absent ('A').  The combined flag is set to
#:  marginal ('M') for any spot that is marginal in the 'Name' column.  This
#:  option is performed automatically on all data.
#:
#:  The script applies cut-off criteria to discard data from weak signals:
#:  spots with < 50% of pixels > 2 SD above median local background signal in
#:  one or both channels are flagged absent, unless one channel shows > 95%
#:  of the pixels > 2 SD above local background.  The SD can be calculated
#:  using only the lower 55% of the pixel intensities (called SD2 in GenePix
#:  Pro), as this measure is less susceptible to being skewed by bright
#:  pixels.  The default cut-offs of 50% and 95% can be changed in the GUI,
#:  and this filtering option can also be switched off.  The script also
#:  provides a quality control report showing the numbers of spots discarded
#:  during the various steps of the data analysis pipeline as well as
#:  information about genes with % CVs of signal ratios > 30% between
#:  replicate spots.
#:
#:  The script also performs a local normalisation using a sliding square
#:  window of spots surrounding each spot.  A user-defined minimum number of
#:  spots is chosen to be used with which to normalise over (default is 400).
#:  The window size default is 16 spots.  This means the square contains
#:  33 x 33 spots (1089) surrounding central spots,
#:  33 x 17 spots (561) surrounding spots at the edge of the array, and
#:  17 x 17 spots (289) surrounding spots in a corner of the array.  Only
#:  spots that are flagged present are used for the normalisation.  Hence,
#:  using a window of 16 means that sometimes, especially for spots close
#:  to the corners of the array, less than 400 spots may actually be used for
#:  the normalisation.  In cases where the block size chosen is small, the
#:  window size is increased up to a user-defined maximum window size
#:  (default is 24) so that at least 600 total spots are in the square.  This
#:  means the block size used with this window change is larger than may be
#:  necessary to optimise the chances of having 400 present spots to use for
#:  normalisation.  This is a heuristic to make the algorithm faster for the
#:  majority of spots, since counting the number of present spots in the
#:  initial square uses a relatively large amount of computational time.  If,
#:  during normalisation, the number of spots is still found to be less than
#:  400, the window is increased further until the maximum window size is
#:  reached.  In these cases, the spots that do use less than 400 spots for
#:  normalisation are reported in the output log file.  The script then
#:  calculates a normalisation factor such that the median signal ratio of
#:  all measurable spots within the square equals 1, and this factor is then
#:  used to scale the signal ratio for the central spot.  The script also
#:  calculates overall Cy3 and Cy5 signals within the square to get a second
#:  normalisation factor to apply to the signals.  This may be a more
#:  accurate approach when large changes occur in signal intensities; these
#:  signal data can be found in the normalised spots file.
#:
#:  The expression data may also been normalised using spikes that can be
#:  added to each of the RNA samples prior to labelling.  For this script to
#:  work, such spikes should be masked with a 'Q'.  There are two primary
#:  approaches that can be used.  i) Use all experimental spots to normalise
#:  all other spots locally as described above, and then apply a correction
#:  factor obtained from the spikes locally normalised in the first round.
#:  This second round can be done either as a global mean ratio of bacterial
#:  controls or it can be done as a second local normalisation round.  Or ii)
#:  use just the spikes to correct the other spots.  Again, this can be done
#:  either globally or locally.  If it is done locally, then additional
#:  parameters are required for the number of spots used in the square and
#:  number of spots required in the initial square.  The same window sizes
#:  are used as when all experimental spots are being used for the local
#:  normalisation.  The advantage of i) over ii) is that by using the
#:  experimental spots rather than the spikes allows for improved monitoring
#:  of the local Cy biases as there are many more of these spots throughout
#:  the array.  The basic assumption underlying the two main methods
#:  described here is that there should be equal values for Cy3 and Cy5 spot
#:  intensities for the spikes, taking into account any observed differences
#:  due to experimental artifacts in labelling or scanning.  A previous
#:  report has suggested that this type of local normalisation, which uses
#:  all local experimental spots, is more robust than if information is
#:  collected just from local spikes (van de Peppel et al., EMBO Reports
#:  2003, 4:387-93).  One advantage of our local normalisation method
#:  compared to the Lowess technique used by van de Peppel et al. (2003) is
#:  that our method is much less susceptible to issues that involve the
#:  skewing of high intensity genes by ribosomal genes; these genes are often
#:  highly down regulated in global responses of cells to large physiological
#:  changes that are observed in stress or starvation.
#:
#:  The signal ratios used for normalisation correspond to the median of all
#:  pixel-by-pixel ratios of signals minus median local background for each
#:  pixel of a given spot (called 'median of ratios' in GenePix Pro).  This
#:  measures ratios more reliably and is less affected by unspecific signals
#:  than the 'ratio of medians'.  In the rare cases where the
#:  'median of ratios' is zero, the 'ratio of medians' is used instead for
#:  data evaluation.  Finally, the script averages the normalised data from
#:  all replicate spots that produced measurable signal ratios of the same
#:  genomic element.  These mean normalised ratios can then be used for
#:  downstream data evaluation.
#:
#############################################################################
#
#  Usage:
#
#:  local_normal.pl
#:   [--help                            ]
#:   [--os          <Boolean>           ]
#:   [--file        <File>              ]
#:   [--filetext    <String>            ]
#:   [--genefile    <File>              ]
#:   [--extgene     <Boolean>           ]
#:   [--window      <Number>            ]
#:   [--maxwin      <Number>            ]
#:   [--blocksize   <Number>            ]
#:   [--blockstart  <Number>            ]
#:   [--spblocksize <Number>            ]
#:   [--spblockst   <Number>            ]
#:   [--round1      <Text>              ]
#:   [--round2      <Text>              ]
#:   [--bgcut       <Boolean>           ]
#:   [--upperbg     <Number>            ]
#:   [--lowerbg     <Number>            ]
#:   [--diamcut     <Number|Boolean>    ]
#:   [--spotnorm    <Boolean>           ]
#:   [--genenorm    <Boolean>           ]
#:   [--genestats   <Boolean>           ]
#:   [--alldata     <Boolean>           ]
#:   [--log         <Boolean>           ]
#:   [--info        <Boolean>           ]
#:
#:    --help        Shows usage.
#:    --os          OS information;
#:                    Windows is true (1), UNIX is false (0; default).
#:    --file        File(s) to be normalised.
#:    --filetext    String to be added to output files.
#:    --genefile    Gene list file.
#:    --extgene     Use external gene list from separate file;
#:                    (default is gene_list.txt in same directory as script).
#:                    If false, use gene list made from input data file.
#:    --window      Size of normalisation window (default is 16).
#:    --maxwin      Maximum size of normalisation window (default is 24).
#:    --blocksize   Size of normalisation block when using experimental spots
#:                    (default is 400).
#:    --blockstart  Initial size of normalisation block when using
#:                    experimental spots (default is 600).
#:                    This is used to account for absent spots
#:                    in the initial square used.
#:    --spblocksize Size of normalisation block when using spikes
#:                    (default is 20).
#:    --spblockst   Initial size of normalisation block when using spikes
#:                    (default is 30).
#:                    This is used to account for absent spots
#:                    in the initial square used.
#:    --round1      Normalisation performed in first round with experimental
#:                    spots; either: all-non-cont, none or not-using.
#:    --round2      Normalisation performed in final round with spikes;
#:                    either: local, global, none or not-using.
#:    --bgcut       Use background SD cut-offs (default is true).
#:    --upperbg     Upper cut-off for background SD (default is 95).
#:    --lowerbg     Lower cut-off for background SD (default is 50).
#:    --diamcut     Use/set spot diameter cut-off (default is 100);
#:                    default is not to use this cut-off.
#:    --spotnorm    Calculate normalised data for all good spots.
#:    --genenorm    Calculate normalised data for all good genes.
#:    --genestats   Prints extra information about normalised data
#:                    for all good genes (default is false).
#:    --alldata     Prints all data for all spots plus normalised data.
#:    --log         Print log information to a file (default is true);
#:                    otherwise print it to stderr.
#:    --info        Print header information to a file.
#:
#:  If no command line parameters are supplied, a graphical user
#:  interface is started instead.
#:
#############################################################################
#
#  Copyright: (C) 2003  Genome Research Limited.
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the:
#    Free Software Foundation, Inc.
#    59 Temple Place - Suite 330
#    Boston, MA 02111-1307, USA.
#
#############################################################################
#
#  Revision History:
#
#  $Author: cjp $
#  $Date: 2006/02/13 14:05:49 $
#  $Revision: 2.7 $
#  $RCSfile: local_normal.pl,v $
#
#  $Log: local_normal.pl,v $
#% Revision 2.7  2006/02/13  14:05:49  cjp
#% Changes to read in old GPR files using column names.
#%
#% Revision 2.6  2005/08/11 09:27:24  cjp
#% Added the comment for previous log!
#%
#% Revision 2.5  2005/08/11 09:19:10  cjp
#% All Data file now includes plate info for ArrayExpress.
#%
#% Revision 2.4  2005/07/06  14:06:19  cjp
#% Gene IDs now have _F at the end of the ID.
#%
#% Revision 2.3  2005/06/30 09:38:09  cjp
#% Samuel's new primer extensions _A1, _A2, etc.
#%
#% Revision 2.2  2005/02/07 11:15:40  cjp
#% Bug fixes for new spike normalisation script.
#%
#% Revision 2.1  2005/01/31  14:11:22  cjp
#% Initial revision.
#%
#% Revision 2.0  2005/01/31  14:00:00  cjp
#% Moved from version 1 to version 2, which involved many changes.
#% Added normalisation methods using spikes and also auto-detection of
#% column names.
#%
#% Revision 1.29  2004/06/30  09:36:15  cjp
#% Fixed behaviour of 'Q', 'T' and 'Y' genes.
#%
#% Revision 1.28  2004/06/29  13:02:48  cjp
#% Made behaviour of 'T' and 'Y' flags identical to 'Q' flags.
#%
#% Revision 1.27  2004/05/11  14:57:18  cjp
#% Changed bacterial controls to use locally normalised ratios of controls
#%   to get a second value for the overall ratio of bacterial controls.
#%
#% Revision 1.26  2004/02/04  12:14:06  cjp
#% Fixed primer names for all data option.
#%
#% Revision 1.25  2004/02/02  15:49:28  cjp
#% To remove primer suffices.
#%
#% Revision 1.24  2004/01/28  13:29:51  cjp
#% Needed to fix gene list plate position info.
#%
#% Revision 1.23  2004/01/28  10:19:37  cjp
#% Hack to deal with 96-well plate positions in new GPR files.
#%
#% Revision 1.22  2003/11/27  16:38:23  cjp
#% Case sensitive bug for some genes fixed.
#%
#% Revision 1.21  2003/11/10  17:19:07  cjp
#% Bug with dongle error message.
#%
#% Revision 1.20  2003/08/29  15:16:16  cjp
#% Changed dongle error message.
#%
#% Revision 1.19  2003/08/19  05:41:17  cjp
#% Changes for GenePix 5.
#%
#% Revision 1.18  2003/08/12  17:32:36  cjp
#% Problem with last fix.  Changed back to previous code
#%   and actually fixed original bug.
#%
#% Revision 1.17  2003/08/12  15:27:19  cjp
#% Fixed bacterial control bug for absent spots.
#%
#% Revision 1.16  2003/08/11  15:38:14  cjp
#% Fixed recursion bug.
#%
#% Revision 1.15  2003/08/01  15:14:19  cjp
#% Fixed bugs with flags.
#%
#% Revision 1.14  2003/08/01  10:46:42  cjp
#% Various bug fixes.  Also changed the way controls are handled.
#%
#% Revision 1.13  2003/07/30  15:05:04  cjp
#% Fixed a few bugs.
#%
#% Revision 1.12  2003/07/15  07:30:01  cjp
#% Counting marginal genes correctly now.
#%
#% Revision 1.11  2003/07/14  13:08:57  cjp
#% Slight changes.
#%
#% Revision 1.10  2003/07/11  15:23:58  cjp
#% Bug with --extgene option.
#%
#% Revision 1.9  2003/07/11  15:16:46  cjp
#% Slight changes to description.
#%
#% Revision 1.8  2003/07/11  12:36:51  cjp
#% Added extra description info.
#%
#% Revision 1.7  2003/07/08  16:04:52  cjp
#% Fixed bug when no genes with high CVs between spots.
#%
#% Revision 1.6  2003/07/04  12:49:33  cjp
#% Changes for windows.
#%
#% Revision 1.5  2003/07/04  08:33:55  cjp
#% Output format changes.
#%
#% Revision 1.4  2003/06/10  14:27:28  cjp
#% Debug of GenePix3 code.
#%
#% Revision 1.3  2003/05/27  09:19:22  cjp
#% New algorithm to make sure we normalise over a minimum number of spots.
#%
#% Revision 1.2  2003/05/16  15:57:09  cjp
#% Added Tk GUI frontend.
#%
#% Revision 1.1  2003/05/12  16:29:02  cjp
#% Initial revision.
#%
#% Revision 1.0  2003/05/06  09:00:00  cjp
#% Started writing script.
#
#############################################################################

require 5.004;


#############################################################################
#  Pragmas
#  -------

use strict;


#############################################################################
#  Modules
#  -------

use Cwd;
#use Data::Dumper;
use Getopt::Long;
use Tk;


#############################################################################
#  Constants
#  ---------

# Version information, the revision number changes automatically
#   whenever the script is checked into CVS or RCS.

use constant VERSION => q$Revision: 2.7 $ =~ /[.\d]+/g;

# Booleans.

use constant FALSE => 0;
use constant TRUE  => 1;

# OS constants.

use constant UNIX    => 0;
use constant WINDOWS => 1;

# Normalisation options.

use constant NON_SPIKES    => 0;
use constant SPIKES        => 1;
use constant ALL_SPOTS     => 2;
use constant NON_CONTROLS  => 3;
use constant NOT_USING     => 4;
use constant NONE          => 5;
use constant ALL_NON_CONT  => 6;
use constant LOCAL         => 7;
use constant GLOBAL        => 8;

# Default values.

use constant DEF_CV_CUT => 30;          # Default % CV cut-off for genes
                                        #   with highly variable spots.

use constant DEF_BACT_MIN => 10;        # Default for minimum no. of
                                        #   bacterial controls to use.

use constant DEF_GENE_LIST_FILE => 'gene_list.txt'; # Default gene list file.

use constant DEF_FILE_TEXT => '';       # Default string to add to
                                        #   output files.

# Default number options.

use constant DEF_WINDOW_SIZE => 16;     # Default for --window option.
use constant DEF_MAXWIN_SIZE => 24;     # Default for --maxwin option.

use constant DEF_BLOCK_SIZE   => 400;   # Default for --blocksize option.
use constant DEF_BLOCKST_SIZE => 600;   # Default for --blockstart option.

use constant DEF_SP_BLOCK_SIZE   => 20; # Default for --spblocksize option.
use constant DEF_SP_BLOCKST_SIZE => 30; # Default for --spblockst option.

use constant DEF_UPPER_BG_CUT_SIZE => 95;   # Default for --upperbg option.
use constant DEF_LOWER_BG_CUT_SIZE => 50;   # Default for --lowerbg option.

use constant DEF_DIAM_CUT_SIZE => 100;  # Default for diameter cut-off.

# Default text options.

use constant DEF_ROUND1 => 'all-non-cont';  # Default for --round1 option.
use constant DEF_ROUND2 => 'none';          # Default for --round2 option.

# Default boolean options.

use constant DEF_EXT_GENE_LIST => FALSE;    # Default for --extgene option.
use constant DEF_BG_CUT_OPT    => TRUE;     # Default for --bgcut option.
use constant DEF_DIAM_CUT_OPT  => FALSE;    # Default for --diamcut option.
use constant DEF_SPOT_NORM_OPT => TRUE;     # Default for --spotnorm option.
use constant DEF_GENE_NORM_OPT => TRUE;     # Default for --genenorm option.
use constant DEF_GENE_STAT_OPT => FALSE;    # Default for --genestats option.
use constant DEF_ALL_DATA_OPT  => TRUE;     # Default for --alldata option.
use constant DEF_LOG_FILE_OPT  => TRUE;     # Default for --log option.
use constant DEF_INFO_FILE_OPT => FALSE;    # Default for --info option.

# gDataArray indices.

use constant INDEX       => 0;      # Index location on the micro-array.
use constant ARRAY_BLOCK => 1;      # Block in micro-array.
use constant BLOCK_COL   => 2;      # Column in block.
use constant BLOCK_ROW   => 3;      # Row in block.
use constant ARRAY_COL   => 4;      # Column in micro-array.
use constant ARRAY_ROW   => 5;      # Row in micro-array.
use constant RATIO       => 6;      # Ratio from GenePix file.
use constant X_POS       => 7;      # X position on micro-array.
use constant Y_POS       => 8;      # Y position on micro-array.
use constant MASK        => 9;      # Mask (or name) from GenePix file.
use constant ID          => 10;     # ID from GenePix file.
use constant FLAG        => 11;     # Integrated flag from GenePix file
                                    #   and mask information.
use constant GPFLAG      => 12;     # GenePix flag.
use constant GSFLAG      => 13;     # GeneSpring flag.
use constant B635SD      => 14;     # % of Cy5 signal pixels > BG 635 + 2SD.
use constant B532SD      => 15;     # % of Cy3 signal pixels > BG 532 + 2SD.
use constant F635SAT     => 16;     # % of saturated Cy5 signal pixels.
use constant F532SAT     => 17;     # % of saturated Cy3 signal pixels.
use constant DIAM        => 18;     # Spot diameter.
use constant MED_RATIO   => 19;     # Ratio of medians.
use constant RATIO_MED   => 20;     # Median of ratios.
use constant SIG_EXPT    => 21;     # Signal1 (experiment).
use constant SIG_CONT    => 22;     # Signal2 (control).
use constant TRUST       => 23;     # Trust is lowest signal in
                                    #   experiment or control.
use constant AVG_LOG_SIG => 24;     # Average log intensity for both signals.
use constant SIG_RATIO   => 25;     # Ratio of signals.

# Parameters from first round of normalisation if 2 rounds performed.
use constant NO_SPOTS0        => 26;    # No. of spots to calculate the
                                        #   normalistion factor.
use constant NORM_RATIO0      => 27;    # Normalised ratio.
use constant NORM_MED_RATIO0  => 28;    # Normalised median of ratios in
                                        #   square.
use constant NORM_SIG_FACTOR0 => 29;    # Normalisation factor from signals.
use constant NORM_SIG_RATIO0  => 30;    # Normalised ratio of signals.
use constant NORM_SIG_EXPT0   => 31;    # Normalised signal for experiment.
use constant FLAG_NORM_IN_0   => 32;    # Flag that spot was normalised.
use constant COUNT_USED_IN_0  => 33;    # Number of times that spot was used
                                        #   for normalisation.

# Parameters from final round of normalisation.
use constant NO_SPOTS1        => 34;    # No. of spots to calculate the
                                        #   normalistion factor.
use constant NORM_RATIO1      => 35;    # Normalised ratio.
use constant NORM_MED_RATIO1  => 36;    # Normalised median of ratios in
                                        #   square.
use constant NORM_SIG_FACTOR1 => 37;    # Normalisation factor from signals.
use constant NORM_SIG_RATIO1  => 38;    # Normalised ratio of signals.
use constant NORM_SIG_EXPT1   => 39;    # Normalised signal for experiment.
use constant FLAG_NORM_IN_1   => 40;    # Flag that spot was normalised.
use constant COUNT_USED_IN_1  => 41;    # Number of times that spot was used
                                        #   for normalisation.

# gGeneHash indices.

use constant NO_VAL     => 0;       # No. of values for this gene.
use constant MARG_FLAG  => 1;       # Flag if this is a marginal gene.
use constant MEAN       => 2;       # Mean value for this gene.
use constant CV         => 3;       # % CV for this gene.
use constant DATA_START => 4;       # Index where normalised ratios begin.


#############################################################################
#  Globals
#  -------

use vars qw(%gOpts $gFileParts
            $gHeader %gPosCols
            $gTk $gDirSep $gCyRatio
            $gDataArray @gGeneList $gGeneHash
            $gColsBlock $gRowsBlock $gBlocksArray
            $gBigX $gBigY);

undef %gOpts;       # Hash of options.

$gFileParts = undef;    # Components of file name being processed.

$gHeader = undef;   # Hash of header information for current file.

undef %gPosCols;    # Positions of columns in GPR file.

$gTk = undef;       # Main Tk window.

$gDirSep = '/';     # Directory separator - '/' for UNIX,
                    #   and '\' for Windows.

$gCyRatio = 0;      # Unknown Cy dye ratio.
                    #   1 is 635/532 (Cy5/Cy3); 2 is 532/635 (Cy3/Cy5).

$gDataArray = undef;    # Array of data for current file.

undef @gGeneList;   # Array of genes on micro-array;
                    #   either specified in a gene list file
                    #   or in the current file.

$gGeneHash = undef; # Hash of data for genes on micro-array
                    #   in current file.

$gColsBlock   = 0;  # No. of columns in a block on the micro-array.
$gRowsBlock   = 0;  # No. of rows in a block on the micro-array.
$gBlocksArray = 0;  # No. of blocks on the micro-array.

$gBigX = 0;         # Largest possible spot number in X direction.
$gBigY = 0;         # Largest possible spot number in Y direction.


#############################################################################
#  setGlobals
#  ----------

sub setGlobals {

  # These global variables are file-specific;
  #   and should be reset for each new GenePix file.

  $gFileParts = undef;

  $gHeader = undef;

  $gCyRatio = 0;

  undef %gPosCols;

  $gDataArray = undef;

  $gGeneHash = undef;

  $gColsBlock   = 0;
  $gRowsBlock   = 0;
  $gBlocksArray = 0;

  $gBigX = 0;
  $gBigY = 0;

  return TRUE;

} # setGlobals


#############################################################################
#  showUsage
#  ---------

sub showUsage {

  print STDERR
    "\n", (split /$gDirSep/, $0)[-1],
    ' (v ' . VERSION . ")\n\n",
    `perl -ne "s/^#://, print if /^#:/" $0`;

  print STDERR
    "ERROR: ", join(' ', @_), "\n"
      if defined $_[0];

  return TRUE;

} # showUsage


#############################################################################
#  showExitMessage
#  ---------------

sub showExitMessage {

  my $text = join ' ', @_;

  my $warn =
    $gTk->Toplevel();

  $warn->title("Warning");

  my %gui_opts = (-text       => $text,
                  -foreground => 'red',
                  -anchor     => 'c');

  my $warn_label =
    $warn->Label(%gui_opts)->pack();

  %gui_opts = (-text       => 'Exit',
               -foreground => 'red',
               -command    => sub { exit; });

  my $warn_button =
    $warn->Button(%gui_opts)->pack();

  $warn->waitWindow;

  return TRUE;

} # showExitMessage


#############################################################################
#  showWarning
#  -----------

sub showWarning {

  my $text = join ' ', @_;

  my $warn =
    $gTk->Toplevel();

  $warn->title("Warning");

  my %gui_opts = (-text       => $text,
                  -foreground => 'red',
                  -anchor     => 'c');

  my $warn_label =
    $warn->Label(%gui_opts)->pack();

  %gui_opts = (-text       => 'Close',
               -foreground => 'red',
               -command    => sub { $warn->withdraw; });

  my $warn_button =
    $warn->Button(%gui_opts)->pack();

  $warn->waitWindow;

  return TRUE;

} # showWarning


#############################################################################
#  Die
#  ---

sub Die {

  my $text = shift;

  if ($gOpts{tk} == TRUE) {

    showExitMessage($text)
      or die $text . "\nCannot show pop-up window.\n";

  } else {

    die $text . "\n";

  }

  return TRUE;

} # Die


#############################################################################
#  Warn
#  ----

sub Warn {

  my $text = shift;

  if ($gOpts{tk} == TRUE) {

    showWarning($text)
      or Die($text);

  } else {

    warn $text . "\n";

  }

  return TRUE;

} # Warn


#############################################################################
#  fileParts
#  ---------

sub fileParts {

  my $file = shift;

  my $disk = '';

  #print "[$gDirSep] [$disk] [$file]\n";

  if ($file =~ /^[A-Z]:/i) {

    ($disk, $file) = $file =~ /^([A-Z]):(.*$)/i;
    $disk .= ':';

  }

  #print "[$gDirSep] [$disk] [$file]\n";

  my @list = split /$gDirSep/, $file;

  my $dir = join $gDirSep, @list[0..$#list-1];

  unless ($dir) {

    $dir  = '.';
    $file = '.' . $gDirSep . $file;

  }

  my $name = $list[-1];

  my @name_list = split /\./, $name;

  my $base = join '.', @name_list[0..$#name_list-1];

  my $ext = $name_list[-1];

  if ($disk =~ /^[A-Z]:$/i) {

    $file = $disk . $file;

  }

  #print "[$file] [$disk] [$dir] [$base] [$ext]\n";

  my $parts = {file => $file,   # Original full file name.
               disk => $disk,   # Windows disk name.
               dir  => $dir,    # Directory part of file (may be a '.').
               base => $base,   # File name without extension.
               ext  => $ext};   # File extension.

  return $parts;

} # fileParts


#############################################################################
#  checkTextFile
#  -------------

sub checkTextFile {

  my $file = shift;

  unless (-f $file) {

    Warn("$file is not a file.");
    return FALSE;

  }

  unless (-r $file) {

    Warn("$file is not a readable file.");
    return FALSE;

  }

  unless (-T $file) {

    Warn("$file is not a text file.");
    return FALSE;

  }

  return TRUE;

} # checkTextFile


#############################################################################
#  getOptError
#  -----------

sub getOptError {

  my $opt = shift;

  my $text = '';

  if ($gOpts{tk} == TRUE) {

    my $opt_string = '^#: +\-\-' . $opt . '\s+';

    $text =
      `perl -ne "s/$opt_string//, tr/\n\r.;//d, print if /$opt_string/" $0`;

  } else {

    $text = "Option --${opt}";

  }

    $text .= ": '$gOpts{$opt}'"
      if defined $gOpts{$opt};

  return $text;

} # getOptError


#############################################################################
#  setBoolOpt
#  ----------

sub setBoolOpt {

  my $opt = shift;
  my $def = shift;

  if (exists $gOpts{$opt} and defined $gOpts{$opt}) {

    $gOpts{$opt} = TRUE
      if $gOpts{$opt} =~ /^[YT]/i;  # Set to TRUE (1) if Yes/True.

    $gOpts{$opt} = FALSE
      if $gOpts{$opt} =~ /^[NF]/i;  # Set to FALSE (0) if No/False.

  } else {

    $gOpts{$opt} = $def;

  }

  $gOpts{$opt} = TRUE               # Must be TRUE (1) -
    unless $gOpts{$opt} == FALSE;   #   if not FALSE (0)!

  return TRUE;

} # setBoolOpt


#############################################################################
#  setNumberOpt
#  ------------

sub setNumberOpt {

  my $opt = shift;
  my $def = shift;

  unless (exists $gOpts{$opt} and defined $gOpts{$opt}) {

    $gOpts{$opt} = $def;

  }

  unless ($gOpts{$opt} =~ /^[\d\.]+$/) {

    my $text = getOptError($opt);

    $text .= ' is not a number.';

    if ($gOpts{tk} == TRUE) {

      showWarning($text)
        or Die($text);

      return FALSE;

    } else {

      showUsage($text)
        or Die($text);

      return FALSE;

    }

  }

  return TRUE;

} # setNumberOpt


#############################################################################
#  setStringOpt
#  ------------

sub setStringOpt {

  my $opt = shift;
  my $def = shift;

  unless (exists $gOpts{$opt} and defined $gOpts{$opt}) {

    $gOpts{$opt} = $def;

  }

  $gOpts{$opt} =~ s/^\s+//;
  $gOpts{$opt} =~ s/\s+$//;

  unless ($gOpts{$opt} =~ /^[A-Z\d\-\_]*$/i) {

    my $text = getOptError($opt);

    (my $char) = $gOpts{$opt} =~ /([^A-Z\d\-\_])/i;

    $text .= ' has an invalid character (' . $char . ') in it.';

    if ($gOpts{tk} == TRUE) {

      showWarning($text)
        or Die($text);

      return FALSE;

    } else {

      showUsage($text)
        or Die($text);

      return FALSE;

    }

  }

  return TRUE;

} # setStringOpt


#############################################################################
#  setTextOpt
#  ----------

sub setTextOpt {

  my $opt = shift;
  my $def = shift;

  if (exists $gOpts{$opt} and defined $gOpts{$opt}) {

    #print "[$opt] [$gOpts{$opt}]\n";

    if ($gOpts{$opt} =~ /^[anslg]/i) {

    #  my %opt_hash = ('non-spikes'    => NON_SPIKES,
    #                  'spikes'        => SPIKES,
    #                  'all-spots'     => ALL_SPOTS,
    #                  'non-cont'      => NON_CONTROLS,
    #                  'not-using'     => NOT_USING,
    #                  'none'          => NONE,
    #                  'all-non-cont'  => ALL_NON_CONT,
    #                  'local'         => LOCAL,
    #                  'global'        => GLOBAL);

    #  my %opt_hash_regex = ('non-spikes'    => '^non[\-_]spikes\$',
    #                        'spikes'        => '^spikes\$',
    #                        'all-spots'     => '^all[\-_]spots\$',
    #                        'non-cont'      => '^non[\-_]cont\$',
    #                        'not-using'     => '^not[\-_]using\$',
    #                        'none'          => '^none\$',
    #                        'all-non-cont'  => '^all[\-_]non[\-_]cont\$',
    #                        'local'         => '^local\$',
    #                        'global'        => '^global\$');

    #SET_TEXT_OPT:
    #  foreach my $key (keys %opt_hash) {

    #    $gOpts{$opt} = $opt_hash{$key}, last SET_TEXT_OPT
    #      if $gOpts{$opt} =~ /$opt_hash_regex{$key}/;

    #  }

      SWITCH: {

        $gOpts{$opt} = NON_SPIKES,    last SWITCH
          if $gOpts{$opt} =~ /^non[\-_]spikes$/i;

        $gOpts{$opt} = SPIKES,        last SWITCH
          if $gOpts{$opt} =~ /^spikes$/i;

        $gOpts{$opt} = ALL_SPOTS,     last SWITCH
          if $gOpts{$opt} =~ /^all[\-_]spots$/i;

        $gOpts{$opt} = NON_CONTROLS,  last SWITCH
          if $gOpts{$opt} =~ /^non[\-_]cont/i;

        $gOpts{$opt} = NOT_USING,     last SWITCH
          if $gOpts{$opt} =~ /^not[\-_]using$/i;

        $gOpts{$opt} = NONE,          last SWITCH
          if $gOpts{$opt} =~ /^none$/i;

        $gOpts{$opt} = ALL_NON_CONT,  last SWITCH
          if $gOpts{$opt} =~ /^all[\-_]non[\-_]cont/i;

        $gOpts{$opt} = LOCAL,         last SWITCH
          if $gOpts{$opt} =~ /^local$/i;

        $gOpts{$opt} = GLOBAL,        last SWITCH
          if $gOpts{$opt} =~ /^global$/i;

      }

    } elsif ($gOpts{$opt} =~ /^\d$/) {

      # Do nothing!

    } else {

      # Default.

      $gOpts{$opt} = NOT_USING;

    }

    #my %opt_hash_print = (NON_SPIKES,    'non-spikes',
    #                      SPIKES,        'spikes',
    #                      ALL_SPOTS,     'all-spots',
    #                      NON_CONTROLS,  'non-cont',
    #                      NOT_USING,     'not-using',
    #                      NONE,          'none',
    #                      ALL_NON_CONT,  'all-non-cont',
    #                      LOCAL,         'local',
    #                      GLOBAL,        'global');

    #print "[$opt] [$gOpts{$opt}] [$opt_hash_print{$gOpts{$opt} }]\n";

  } else {

    $gOpts{$opt} = $def;

  }

  return TRUE;

} # setTextOpt


#############################################################################
#  resetTextOpt
#  ------------

sub resetTextOpt {

  my $opt = shift;

  my %opt_hash = (NON_SPIKES,    'non-spikes',
                  SPIKES,        'spikes',
                  ALL_SPOTS,     'all-spots',
                  NON_CONTROLS,  'non-cont',
                  NOT_USING,     'not-using',
                  NONE,          'none',
                  ALL_NON_CONT,  'all-non-cont',
                  LOCAL,         'local',
                  GLOBAL,        'global');

  $gOpts{$opt} = $opt_hash{$gOpts{$opt} };

  return TRUE;

} # resetTextOpt


#############################################################################
#  resetTextOpts
#  -------------

sub resetTextOpts {

  my @opts = qw(round1
                round2);

 OPT:
  foreach my $opt (@opts) {

    my $text = getOptError($opt);

    $text .= ' is an invalid value for this option.';

    resetTextOpt($opt)
      or Die(getOptError($opt) . ' is an invalid value for this option.');

  }

  return TRUE;

} # resetTextOpts


#############################################################################
#  setOpts
#  -------

sub setOpts {

  my $opt = 'os';

  my $text = getOptError($opt);

  $text .= ' cannot be set as a boolean value.';

  setBoolOpt($opt, UNIX)
    or Die($text);

  if ($gOpts{$opt} == WINDOWS) {
    $gDirSep = '\\\\';

  } elsif ($gOpts{$opt} == UNIX) {
    $gDirSep = '/';

  } else {
    $gDirSep = '/';

  }

  # Windows separator '\\\\' does not work, so always use '/'.
  $gDirSep = '/';

  #print "[$gOpts{$opt}] [$gDirSep]\n";

  $opt = 'genefile';

  if (exists $gOpts{extgene} and
      defined $gOpts{extgene} and
      $gOpts{extgene}) {

    if (exists $gOpts{$opt} and defined $gOpts{$opt}) {

      # It has been specified by the user.

    } else {

      # Use default gene list file.

      #my $file_base = $0;

      #$file_base =~ s/^([^$gDirSep]*)$gDirSep.*$/$1/;

      #$gOpts{$opt} = $file_base . $gDirSep . DEF_GENE_LIST_FILE;

      #if ($gOpts{os} == WINDOWS) {

      #  $gOpts{$opt} =~ s/\//\\/g;

      #}

      my $file_parts = fileParts($0);

      $gOpts{$opt} = $file_parts->{disk} .
                     $file_parts->{dir}  . $gDirSep .
                     DEF_GENE_LIST_FILE;

      #print "[$gOpts{$opt}]\n";

    }

    checkTextFile($gOpts{$opt})
      or Die("Cannot check file: $gOpts{$opt}.");

  }

  $opt = 'extgene';

  setBoolOpt($opt, DEF_EXT_GENE_LIST)
    or Die($text);

  $opt = 'diamcut';

  if (exists $gOpts{$opt} and defined $gOpts{$opt}) {

    if ($gOpts{$opt} =~ /^[YTNF]/i) {

      my $text = getOptError($opt);

      $text .= ' cannot be set as a boolean value.';

      setBoolOpt($opt, DEF_DIAM_CUT_OPT)
        or Die($text);

      $gOpts{$opt} = DEF_DIAM_CUT_SIZE
        if $gOpts{$opt} == TRUE;

    } else {

      my $text = getOptError($opt);

      $text .= ' is not a number.';

      setNumberOpt($opt, DEF_DIAM_CUT_SIZE)
        or Die($text);

    }

  } else {

    $gOpts{$opt} = DEF_DIAM_CUT_OPT;

    $gOpts{$opt} = DEF_DIAM_CUT_SIZE
      if $gOpts{$opt} == TRUE;

  }

  my @opts = qw(filetext
                window
                maxwin
                blocksize
                blockstart
                spblocksize
                spblockst
                round1
                round2
                bgcut
                upperbg
                lowerbg
                spotnorm
                genenorm
                genestats
                alldata
                log
                info);

  my $bool = \&setBoolOpt;
  my $num  = \&setNumberOpt;
  my $txt  = \&setTextOpt;
  my $str  = \&setStringOpt;

  my %opt_set = (filetext    => $str,
                 window      => $num,
                 maxwin      => $num,
                 blocksize   => $num,
                 blockstart  => $num,
                 spblocksize => $num,
                 spblockst   => $num,
                 round1      => $txt,
                 round2      => $txt,
                 bgcut       => $bool,
                 upperbg     => $num,
                 lowerbg     => $num,
                 spotnorm    => $bool,
                 genenorm    => $bool,
                 genestats   => $bool,
                 alldata     => $bool,
                 log         => $bool,
                 info        => $bool);

  my %opt_def = (filetext    => DEF_FILE_TEXT,
                 window      => DEF_WINDOW_SIZE,
                 maxwin      => DEF_MAXWIN_SIZE,
                 blocksize   => DEF_BLOCK_SIZE,
                 blockstart  => DEF_BLOCKST_SIZE,
                 spblocksize => DEF_SP_BLOCK_SIZE,
                 spblockst   => DEF_SP_BLOCKST_SIZE,
                 round1      => DEF_ROUND1,
                 round2      => DEF_ROUND2,
                 bgcut       => DEF_BG_CUT_OPT,
                 upperbg     => DEF_UPPER_BG_CUT_SIZE,
                 lowerbg     => DEF_LOWER_BG_CUT_SIZE,
                 spotnorm    => DEF_SPOT_NORM_OPT,
                 genenorm    => DEF_GENE_NORM_OPT,
                 genestats   => DEF_GENE_STAT_OPT,
                 alldata     => DEF_ALL_DATA_OPT,
                 log         => DEF_LOG_FILE_OPT,
                 info        => DEF_INFO_FILE_OPT);

 OPT:
  foreach my $opt (@opts) {

    my $text = getOptError($opt);

    $text .= ' is an invalid value for this option.';

    $opt_set{$opt}->($opt, $opt_def{$opt})
      or Die(getOptError($opt) . ' is an invalid value for this option.');

  }

  return TRUE;

} # setOpts


#############################################################################
#  openLog
#  -------

sub openLog {

  my $file = shift;

  my $file_log = '';

  #local *LOG;

  my $fh = undef;       # $fh is file handle for log file.

  if ($gOpts{log} == TRUE) {

    $file_log = $gFileParts->{disk} .
                $gFileParts->{dir}  . $gDirSep .
                $gFileParts->{base} . '_';

    my $extra_ul = '';

    $extra_ul = $gOpts{filetext} ? '_' : '';

    $file_log .=  $gOpts{filetext} . $extra_ul . 'log.txt';

    open LOG, '>' . $file_log
      or Die("Cannot create file $file_log: $!");

    $fh = \*LOG;

    print STDERR 'Log file: ' . $file_log . "\n\n";

  } elsif ($gOpts{log} == FALSE) {

    # Only do this code once!

    $gOpts{log} = -1;

    $fh = \*STDERR;

    print $fh "Printing log information to stderr.\n\n";

  } elsif ($gOpts{log} == -1) {

    # Do nothing!

  }

  return $fh;

} # openLog


#############################################################################
#  getTimeDate
#  -----------

sub getTimeDate {

  my @array = localtime(time);

  my %month_hash = (1  => 'Jan',
                    2  => 'Feb',
                    3  => 'Mar',
                    4  => 'Apr',
                    5  => 'May',
                    6  => 'Jun',
                    7  => 'Jul',
                    8  => 'Aug',
                    9  => 'Sep',
                    10 => 'Oct',
                    11 => 'Nov',
                    12 => 'Dec');

  my $time_date = sprintf "%02d:%02d on %02d %03s %02d",
    $array[2], $array[1], $array[3], $month_hash{1+$array[4]},
    1900 + $array[5];

  return $time_date

} # getTimeDate


#############################################################################
#  trimLine
#  --------

sub trimLine {

  my $line = shift;

  chomp $line;

  $line =~ tr/\"\r//d;

  #my @array = split /\t/, $line;

  #s/^\s+//, s/\s+$// foreach @array;

  #$line = join "\t", @array;

  return $line;

} # trimLine


#############################################################################
#  readGeneFile
#  ------------

sub readGeneFile {

  open GENE, $gOpts{genefile}
    or Die("Cannot open file $gOpts{genefile}: $!");

  while (defined (my $line = <GENE>) ) {

    push @gGeneList, lc trimLine($line);

  }

  close GENE
    or Die("Cannot close file $gOpts{genefile}: $!");

  return TRUE;

} # readGeneFile


#############################################################################
#  readColumns
#  -----------

sub readColumns {

  my $line = shift;

  my @cols = split /\t/, $line;

  #print "[@cols]\n";

  my $i = 0;

  foreach my $col (@cols) {
    $gPosCols{$col} = $i;
    $i++;
  }

  #foreach my $col (sort keys %gPosCols) {
  #  print "[$col] [$gPosCols{$col}]\n";
  #}

  return TRUE;

} # readColumns


#############################################################################
#  initialReadFile
#  ---------------

sub initialReadFile {

  my $fh = shift;

  my $file = $gFileParts->{file};

  open FILE, $file
    or Die("Cannot open file $file: $!");

  my $gene_flag = FALSE;

  # Only need a gene list if gene averaged data required.

  if ($gOpts{genenorm} == TRUE) {

    unless ($gOpts{extgene}) {

      undef @gGeneList;

      my $string = "Using gene list from genes in file: $file.\n\n";

      print $fh $string
        if $gOpts{log} == TRUE;

      print STDERR $string;

      $gene_flag = TRUE;

    } else {

      my $string = "Using gene list file: $gOpts{genefile}.\n\n";

      print $fh $string
        if $gOpts{log} == TRUE;

      print STDERR $string;

    }

  }

  my $header_flag = FALSE;

  my $old_col = 0;
  my $old_row = 0;
  my $max_x   = 0;

  my $count = 0;

 WHILE:
  while (defined (my $line = <FILE>) ) {

    $line = trimLine($line);

    if ($line =~ /^Block/) {

      readColumns($line)
        or Die("Cannot read columns in this file.");

      $header_flag = TRUE;

      next WHILE;

    }

    next WHILE
      unless $header_flag == TRUE;

    $count++;

    my @array = split /\t/, $line;

    my $block = $array[$gPosCols{Block}];
    my $col   = $array[$gPosCols{Column}];
    my $row   = $array[$gPosCols{Row}];
    my $x_pos = $array[$gPosCols{X}];

    #print "[$block] [$col] [$row] [$x_pos]\n";

    if ($gene_flag == TRUE) {

      my $id = $array[$gPosCols{ID}];

      $id =~ s/\s+$//;

      # Hack to remove 96-well plate positions in new GPR files.

      if ($id =~ /^\d+\-[A-H]\d+_/) {

        $id =~ s/^\d+\-[A-H]\d+_//;

      }

      if ($id =~ /_[FR]$/) {

        $id =~ s/_[FR]$//;

      }

      # To remove primer suffices.

      if ($id =~ /_\d+$/) {

        $id =~ s/_\d+$//;

      }

      push @gGeneList, lc $id;

    }

    #print "COL: $col\t$old_col\t$gColsBlock\n";

    if ($col > $old_col) {

      $old_col = $col;

    } elsif ($col < $old_col) {

      $gColsBlock = $old_col
        if $gColsBlock == 0;

    } elsif ($gColsBlock and
             ($col > $gColsBlock or $old_col > $gColsBlock) ) {

      Die("The number of columns per block changes in the GenePix file.");

    }

    #print "ROW: $row\t$old_row\t$gRowsBlock\n";

    if ($row > $old_row) {

      $old_row = $row;

    } elsif ($row < $old_row) {

      $gRowsBlock = $old_row
        if $gRowsBlock == 0;

    } elsif ($gRowsBlock and
             ($row > $gRowsBlock or $old_row > $gRowsBlock) ) {

      Die("The number of rows per block changes in the GenePix file.");

    }

    if ($col == $gColsBlock and $row == $gRowsBlock) {

      #print "X: $x_pos\t$max_x\t$gBlocksArray\n";

      if ($x_pos > $max_x) {

        $max_x = $x_pos;

      } elsif ($x_pos < $max_x) {

        #print "$block\n";

        $gBlocksArray = $block - 1
          if $gBlocksArray == 0;

        #my $mod = ($block - 1)%$gBlocksArray;

        #print "[$mod]\n";

        $max_x = 0;

      } elsif ($gBlocksArray and ( ($block - 1)%$gBlocksArray != 0) ) {

        Die("The number of blocks per array changes in the GenePix file.");

      }

    }

  }

  my $string =
    "Analysis performed without a dongle in file $file.\n" .
    "Please re-analyse with a dongle in your computer.\n";

  Die($string)
    if $count == 100 or $count == 1000;

  $string = "Cannot calculate columns per block in file $file.\n";

  Die($string)
    if $gColsBlock == 0;

  $string = "Columns per block:\t" . $gColsBlock . ".\n";

  print $fh $string
    if $gOpts{log} == TRUE;

  print STDERR $string;

  Die("Cannot calculate rows per block in file $file.\n")
    if $gRowsBlock == 0;

  $string = "Rows per block:\t\t" . $gRowsBlock . ".\n";

  print $fh $string
    if $gOpts{log} == TRUE;

  print STDERR $string;

  Die("Cannot calculate blocks per array in file $file.\n")
    if $gBlocksArray == 0;

  $string = "Blocks per array:\t" . $gBlocksArray . ".\n\n";

  print $fh $string
    if $gOpts{log} == TRUE;

  print STDERR $string;

  close FILE
    or Die("Cannot close file $file: $!");

  return TRUE;

} # initialReadFile


#############################################################################
#  readHeader
#  ----------

sub readHeader {

  my $fh   = shift;
  my $line = shift;

  my $sd2_flag = TRUE;

  if ($line =~ /^Block\s+/) {

    $line =~ s/Rgn R.( ?)/Rgn R2$1/g;   # Replaces unreadable sign with 2.

    $line =~ s/^(Block)\s+/Columns=$1\t/;

    if ($line =~ /SD2/) {

      $sd2_flag = TRUE;

      print $fh
        "Using SD2 for background standard deviations.\n";

    } else {

      $sd2_flag = FALSE;

      print $fh
        "Using SD for background standard deviations.\n";

    }

  }

  if ($line =~ /^RatioFormulation=/) {

    $line =~ s/^RatioFormulation=/RatioFormulations=/;

  }

  #my $file = $gFileParts->{file};

  # Could check header line is correct here.

  my ($title, @desc) = split /[=\t]+/, $line;

  my $desc = join '; ', @desc;

  $gHeader->{$title} = $desc;

  return $sd2_flag;

} # readHeader


#############################################################################
#  getHeaderInfo
#  -------------

sub getHeaderInfo {

  my $fh = shift;

  my $gp_vers = 3;

  if ($gHeader->{RatioFormulations} =~ m!635( nm)?/532( nm)?!) {

    $gCyRatio = 1;

    print $fh
      "Ratio: 635/532.\n";

  } elsif ($gHeader->{RatioFormulations} =~ m!532( nm)?/635( nm)?!) {

    $gCyRatio = 2;

    print $fh
      "Ratio: 532/635.\n";

  }

  #if ($gHeader->{Creator} =~ /^GenePix Pro 4\./) {

  #  $gp_vers = 4;

  #  print $fh
  #    "File type: GenePix version 4.\n";

  #} elsif ($gHeader->{Creator} =~ /^GenePix Pro 3\./) {

  #  $gp_vers = 3;

  #  print $fh
  #    "File type: GenePix version 3.\n";

  #} else {

  #  $gp_vers = 3;

  #  print $fh
  #    "File type: unknown GenePix version; using default of version 3.\n";

  #}

  if ($gHeader->{Type} =~ /^GenePix Results 3/) {

    $gp_vers = 5;

    print $fh
      "File type: GenePix version 5.\n";

  } elsif ($gHeader->{Type} =~ /^GenePix Results 2/) {

    $gp_vers = 4;

    print $fh
      "File type: GenePix version 4.\n";

  } elsif ($gHeader->{Type} =~ /^GenePix Results 1/) {

    $gp_vers = 3;

    print $fh
      "File type: GenePix version 3.\n";

  } else {

    $gp_vers = 3;

    print $fh
      "File type: unknown GenePix version; using default of version 3.\n";

  }

  print $fh "\n";

  return $gp_vers

} # getHeaderInfo


#############################################################################
#  printHeader
#  -----------

sub printHeader {

  my $file_header = $gFileParts->{disk} .
                    $gFileParts->{dir}  . $gDirSep .
                    $gFileParts->{base} . '_';

  my $extra_ul = '';

  $extra_ul = $gOpts{filetext} ? '_' : '';

  $file_header .=  $gOpts{filetext} . $extra_ul . 'header.txt';

  open HEADER, '>' . $file_header
    or Die("Cannot create file $file_header: $!");

  foreach my $key (sort keys %{$gHeader}) {
    print HEADER $key . ":\t" . $gHeader->{$key} . "\n";
  }

  close HEADER
    or Die("Cannot close file $file_header: $!");

  return TRUE;

} # printHeader


#############################################################################
#  versGPLines
#  -----------

sub versGPLines {

  my $array   = shift;
  my $gp_vers = shift;

  my $b635sd = 0;
  my $b532sd = 0;

  my $f635sat = 0;
  my $f532sat = 0;

  # $ratio is median of ratios, or,
  #   if this is zero, then ratio of medians.

  my $med_ratio = 0;
  my $ratio_med = 0;

  # These are the values of the median backgrounds;
  #   also called 'foreground median' - 'background'.

  my $sig_expt = 0;
  my $sig_cont = 0;

  my $flag = '';

  # Positions of spot variables in the array are different
  #   between versions 3 and 4 of GenePix.

  my $b635sd_pos = 0;
  my $b532sd_pos = 0;

  my $f635sat_pos = 0;
  my $f532sat_pos = 0;

  my $med_ratio_pos = 0;
  my $ratio_med_pos = 0;

  my $flag_pos = 0;

  my $sig_a_pos = 0;
  my $sig_b_pos = 0;

  # Could check for correct size of @{$array} here
  #   between different versions of GenePix.

  if ($gp_vers == 3) {

    $b635sd_pos = 15;
    $b532sd_pos = 24;

    $f635sat_pos = 16;
    $f532sat_pos = 25;

    $med_ratio_pos = 28;
    $ratio_med_pos = 26;

    $flag_pos = 42;

    $sig_a_pos = 38;    # Cy5.
    $sig_b_pos = 39;    # Cy3.

  } elsif ($gp_vers == 4) {

    $b635sd_pos = 15;
    $b532sd_pos = 24;

    $f635sat_pos = 16;
    $f532sat_pos = 25;

    $med_ratio_pos = 46;
    $ratio_med_pos = 44;

    $flag_pos = 80;

    $sig_a_pos = 72;    # Cy5.
    $sig_b_pos = 73;    # Cy3.

  } elsif ($gp_vers == 5) {

    $b635sd_pos = 18;
    $b532sd_pos = 30;

    $f635sat_pos = 19;
    $f532sat_pos = 31;

    $med_ratio_pos = 34;
    $ratio_med_pos = 32;

    $flag_pos = 53;

    $sig_a_pos = 45;    # Cy5.
    $sig_b_pos = 46;    # Cy3.

  }

  $b635sd = $array->[$b635sd_pos];
  $b532sd = $array->[$b532sd_pos];

  $f635sat = $array->[$f635sat_pos];
  $f532sat = $array->[$f532sat_pos];

  $med_ratio = $array->[$med_ratio_pos];

  $ratio_med = $array->[$ratio_med_pos];

  $flag = $array->[$flag_pos];

  if ($gCyRatio == 1) {

    $sig_expt = $array->[$sig_a_pos];
    $sig_cont = $array->[$sig_b_pos];

  } elsif ($gCyRatio == 2) {

    $sig_expt = $array->[$sig_b_pos];
    $sig_cont = $array->[$sig_a_pos];

  }

  return ($b635sd, $b532sd, $f635sat, $f532sat,
          $med_ratio, $ratio_med, $flag, $sig_expt, $sig_cont);

} # versGPLines


#############################################################################
#  versGPLinesNew
#  --------------

sub versGPLinesNew {

  my $array    = shift;
  my $gp_vers  = shift;
  my $sd2_flag = shift;

  my $b635sd = 0;
  my $b532sd = 0;

  if ($sd2_flag == TRUE) {

    $b635sd = $array->[$gPosCols{'% > B635+2SD2'}];
    $b532sd = $array->[$gPosCols{'% > B532+2SD2'}];

  } else {

    $b635sd = $array->[$gPosCols{'% > B635+2SD'}];
    $b532sd = $array->[$gPosCols{'% > B532+2SD'}];

  }

  my $f635sat = $array->[$gPosCols{'F635 % Sat.'}];
  my $f532sat = $array->[$gPosCols{'F532 % Sat.'}];

  # $ratio is median of ratios, or,
  #   if this is zero, then ratio of medians.

  my $med_ratio = 0;
  my $ratio_med = 0;

  # These are the values of the median backgrounds;
  #   also called 'foreground median' - 'background'.

  my $sig_expt = 0;
  my $sig_cont = 0;

  my $flag = $array->[$gPosCols{Flags}];

  # Positions of spot variables in the array are different
  #   between versions 3 and 4 of GenePix.

  if ($gCyRatio == 1) {

    if ($gp_vers == 3) {

      $med_ratio = $array->[$gPosCols{'Median of Ratios'}];
      $ratio_med = $array->[$gPosCols{'Ratio of Medians'}];

    } else {

      $med_ratio = $array->[$gPosCols{'Median of Ratios (635/532)'}];
      $ratio_med = $array->[$gPosCols{'Ratio of Medians (635/532)'}];

    }

    # Was $array->[$sig_a_pos].
    $sig_expt = $array->[$gPosCols{'F635 Median - B635'}];

    # Was $array->[$sig_b_pos].
    $sig_cont = $array->[$gPosCols{'F532 Median - B532'}];

  } elsif ($gCyRatio == 2) {

    if ($gp_vers == 3) {

      $med_ratio = $array->[$gPosCols{'Median of Ratios'}];
      $ratio_med = $array->[$gPosCols{'Ratio of Medians'}];

    } else {

      $med_ratio = $array->[$gPosCols{'Median of Ratios (532/635)'}];
      $ratio_med = $array->[$gPosCols{'Ratio of Medians (532/635)'}];

    }

    # Was $array->[$sig_b_pos].
    $sig_expt = $array->[$gPosCols{'F532 Median - B532'}];

    # Was $array->[$sig_a_pos].
    $sig_cont = $array->[$gPosCols{'F635 Median - B635'}];

  }

  return ($b635sd, $b532sd, $f635sat, $f532sat,
          $med_ratio, $ratio_med, $flag, $sig_expt, $sig_cont);

} # versGPLinesNew


#############################################################################
#  readLine
#  --------

sub readLine {

  my $line     = shift;
  my $sd2_flag = shift;
  my $gp_vers  = shift;

  my @array = split /\t/, $line;

  my $block = $array[$gPosCols{Block}];     # Was $array[0].

  my $block_col = $array[$gPosCols{Column}];    # Column number in the block;
                                                #   was array[1].
  my $block_row = $array[$gPosCols{Row}];       # Row number in the block;
                                                #   was array[2].

  my $mask = $array[$gPosCols{Name}];       # Was array[3].
  my $id   = $array[$gPosCols{ID}];         # Was array[4].

  $mask = 'P'
    unless $mask =~ /^[AMQYT]$/;

  $id =~ s/\s+$//;

  # Hack to remove 96-well plate positions in new GPR files.

  if ($id =~ /^\d+\-[A-H]\d+_/) {

    $id =~ s/^\d+\-[A-H]\d+_//;

  }

  if ($id =~ /_[FR]$/) {

    $id =~ s/_[FR]$//;

  }

  #print "[$id] ";

  # To remove primer suffices.

  if ($id =~ /_\d+$/) {

    $id =~ s/_\d+$//;

  }

  #print "[$id]\n";

  my $x_pos = $array[$gPosCols{X}];     # Was $array[5].
  my $y_pos = $array[$gPosCols{Y}];     # Was $array[6].

  my $diam = $array[$gPosCols{'Dia.'}]; # Was $array[7].

  # Calculate absolute column, row and index position
  #   (within the micro-array).

  my $array_col =
    ( ($block - 1)%$gBlocksArray) * $gColsBlock + $block_col;

  my $array_row =
    (int( ($block - 1)/$gBlocksArray) ) * $gRowsBlock + $block_row;

  # - 1 at the end of the next equation so that array starts at 0, not 1.

  my $index =
    ($array_row - 1) * $gBlocksArray * $gColsBlock + $array_col - 1;

  #print "[$array_col] [$array_row] [$index]\n";

  # These values are in different positions of the GenePix file
  #   for different versions of the software.

  #my ($b635sd, $b532sd, $f635sat, $f532sat,
  #    $med_ratio, $ratio_med, $flag, $sig_expt, $sig_cont)
  #  = versGPLines(\@array, $gp_vers);

  # New version looks at actual column headings using %gPosCols.

  my ($b635sd, $b532sd, $f635sat, $f532sat,
      $med_ratio, $ratio_med, $flag, $sig_expt, $sig_cont)
    = versGPLinesNew(\@array, $gp_vers, $sd2_flag);

  my $ratio = $med_ratio;

  $ratio = $ratio_med
    if $ratio == 0;

  my $trust = 0;

  $trust = ($sig_expt < $sig_cont) ? $sig_expt : $sig_cont;

  my $avg_log = 0;

  $avg_log = (log $sig_expt + log $sig_cont)/2
    if ($sig_expt > 0 and $sig_cont > 0);

  my $data = undef;

  $data->[INDEX]       = $index;    # Index location on the micro-array.
  $data->[ARRAY_BLOCK] = $block;    # Block in micro-array.
  $data->[BLOCK_COL]   = $block_col;    # Column in block.
  $data->[BLOCK_ROW]   = $block_row;    # Row in block.
  $data->[ARRAY_COL]   = $array_col;    # Column in micro-array.
  $data->[ARRAY_ROW]   = $array_row;    # Row in micro-array.
  $data->[RATIO]       = $ratio;    # Ratio from GenePix file.
  $data->[X_POS]       = $x_pos;    # X position on micro-array.
  $data->[Y_POS]       = $y_pos;    # Y position on micro-array.
  $data->[MASK]        = $mask;     # Mask (or name) from GenePix file.
  $data->[ID]          = $id;       # ID from GenePix file.
  $data->[FLAG]        = 'A';       # Integrated flag from GenePix file
                                    #   and mask information.
  $data->[GPFLAG]      = $flag;     # GenePix flag.
  $data->[GSFLAG]      = 'A';       # GeneSpring flag.
  $data->[B635SD]      = $b635sd;   # % of Cy5 signal pixels > BG 635 + 2SD.
  $data->[B532SD]      = $b532sd;   # % of Cy3 signal pixels > BG 532 + 2SD.
  $data->[F635SAT]     = $f635sat;  # % of saturated Cy5 signal pixels.
  $data->[F532SAT]     = $f532sat;  # % of saturated Cy3 signal pixels.
  $data->[DIAM]        = $diam;     # Spot diameter.
  $data->[MED_RATIO]   = $med_ratio;    # Median of ratios.
  $data->[RATIO_MED]   = $ratio_med;    # Ratio of medians.
  $data->[SIG_EXPT]    = $sig_expt; # Signal1 (experiment).
  $data->[SIG_CONT]    = $sig_cont; # Signal2 (control).
  $data->[TRUST]       = $trust;    # Trust is lowest signal in
                                    #   experiment or control.
  $data->[AVG_LOG_SIG] = $avg_log;  # Average log intensity
                                    #   for both signals.
  $data->[SIG_RATIO]   = 0;         # Ratio of signals;

  # Parameters from first round of normalisation if 2 rounds performed.
  $data->[NO_SPOTS0]        = 0;    # No. spots to calculate the
                                    #   normalisation factor.
  $data->[NORM_RATIO0]      = 0;    # Normalised ratio.
  $data->[NORM_MED_RATIO0]  = 0;    # Normalised median of ratios in
                                    #   square.
  $data->[NORM_SIG_FACTOR0] = 0;    # Normalisation factor from signals.
  $data->[NORM_SIG_RATIO0]  = 0;    # Normalised ratio of signals.
  $data->[NORM_SIG_EXPT0]   = 0;    # Normalised signal for experiment.
  $data->[FLAG_NORM_IN_0]   = FALSE;    # Flag that spot was normalised.
  $data->[COUNT_USED_IN_0]  = 0;        # Number of times that spot was used
                                        #   for normalisation.

  # Parameters from final round of normalisation.
  $data->[NO_SPOTS1]        = 0;    # No. spots to calculate the
                                    #   normalisation factor.
  $data->[NORM_RATIO1]      = 0;    # Normalised ratio.
  $data->[NORM_MED_RATIO1]  = 0;    # Normalised median of ratios in
                                    #   square.
  $data->[NORM_SIG_FACTOR1] = 0;    # Normalisation factor from signals.
  $data->[NORM_SIG_RATIO1]  = 0;    # Normalised ratio of signals.
  $data->[NORM_SIG_EXPT1]   = 0;    # Normalised signal for experiment.
  $data->[FLAG_NORM_IN_1]   = FALSE;    # Flag that spot was normalised.
  $data->[COUNT_USED_IN_1]  = 0;        # Number of times that spot was used
                                        #   for normalisation.

  $gDataArray->[$index] = $data;

  return TRUE;

} # readLine


#############################################################################
#  readFile
#  --------

sub readFile {

  my $fh = shift;

  my $file = $gFileParts->{file};

  open FILE, $file
    or Die("Cannot open file $file: $!");

  my $header_flag = FALSE;

  my $sd2_flag = TRUE;
  my $gp_vers  = 3;

  $gHeader    = undef;
  $gDataArray = undef;

 LINE:
  while (defined (my $line = <FILE>) ) {

    $line = trimLine($line);

    if ($header_flag == FALSE) {

      # $header_flag is set to TRUE when at the end of the header.

      $header_flag = TRUE
        if $line =~ /^Block\s+/;

      $sd2_flag = readHeader($fh, $line);

      if ($header_flag == TRUE) {

        # At the end of the header, get information about
        #   Cy ratio and GenePix version

        $gp_vers = getHeaderInfo($fh);

        if ($gOpts{info} == TRUE) {

          # Print header information to a file.

          print STDERR "Printing header information to a file.\n\n";

          printHeader()
            or Die("Cannot print header information for file $file.");

        }

      }

      next LINE;

    }

    unless ($header_flag == TRUE) {
      Die("Cannot read GenePix file $file correctly.");
    }

    # Get relevant data for each spot.

    readLine($line, $sd2_flag, $gp_vers)
      or Die("Cannot read data for line $line in file $file.");

  }

  close FILE
    or Die("Cannot close file $file: $!");

  return TRUE;

} # readFile


#############################################################################
#  calcDimensions
#  --------------

sub calcDimensions {

  $gBigX = $gColsBlock * $gBlocksArray;

  $gBigY = ($#{$gDataArray} + 1)/$gBigX;

  #print "[$gBigX] [$gBigY]\n";

  return TRUE;

} # calcDimensions


#############################################################################
#  readData
#  --------

sub readData {

  my $file = shift;
  my $fh   = shift;

  # Calculate the number of columns per block on the micro-array.

  # Also get a gene list from GenePix file if needed!

  initialReadFile($fh)
    or Die("Cannot pre-parse file $file.");

  # Read data into two global hashes: $gHeader and $gDataArray.

  readFile($fh)
    or Die("Cannot parse file $file.");

  calcDimensions()
    or Die('Cannot calculate dimensions of ' .
           "the micro-array in file $file.");

  return TRUE;

} # readData


#############################################################################
#  zeroData
#  --------

sub zeroData {

  my $data  = shift;
  my $index = shift;

  $data->[INDEX]       = $index;
  $data->[ARRAY_BLOCK] = 0;
  $data->[BLOCK_COL]   = 0;
  $data->[BLOCK_ROW]   = 0;
  $data->[ARRAY_COL]   = 0;
  $data->[ARRAY_ROW]   = 0;
  $data->[RATIO]       = 0;
  $data->[X_POS]       = 0;
  $data->[Y_POS]       = 0;
  $data->[MASK]        = 'A';
  $data->[ID]          = '';
  $data->[FLAG]        = 'A';
  $data->[GPFLAG]      = 0;
  $data->[GSFLAG]      = 'A';
  $data->[B635SD]      = 0;
  $data->[B532SD]      = 0;
  $data->[F635SAT]     = 0;
  $data->[F532SAT]     = 0;
  $data->[DIAM]        = 0;
  $data->[MED_RATIO]   = 0;
  $data->[RATIO_MED]   = 0;
  $data->[SIG_EXPT]    = 0;
  $data->[SIG_CONT]    = 0;
  $data->[TRUST]       = 0;
  $data->[AVG_LOG_SIG] = 0;
  $data->[SIG_RATIO]   = 0;

  $data->[NO_SPOTS0]        = 0;
  $data->[NORM_RATIO0]      = 0;
  $data->[NORM_MED_RATIO0]  = 0;
  $data->[NORM_SIG_FACTOR0] = 0;
  $data->[NORM_SIG_RATIO0]  = 0;
  $data->[NORM_SIG_EXPT0]   = 0;
  $data->[FLAG_NORM_IN_0]   = FALSE;
  $data->[COUNT_USED_IN_0]  = 0;

  $data->[NO_SPOTS1]        = 0;
  $data->[NORM_RATIO1]      = 0;
  $data->[NORM_MED_RATIO1]  = 0;
  $data->[NORM_SIG_FACTOR1] = 0;
  $data->[NORM_SIG_RATIO1]  = 0;
  $data->[NORM_SIG_EXPT1]   = 0;
  $data->[FLAG_NORM_IN_1]   = FALSE;
  $data->[COUNT_USED_IN_1]  = 0;

  return $data;

} # zeroData


#############################################################################
#  maskData
#  --------

sub maskData {

  my $fh = shift;

  my $count = 0;

  my $bad_spots        = 0;
  my $empty_spots      = 0;
  my $marginal_spots   = 0;
  my $masked_spots     = 0;
  my $negative_spots   = 0;
  my $not_found_spots  = 0;
  my $zero_ratio_spots = 0;

 MASK:
  foreach my $index (@{$gDataArray}) {

    unless (defined $index) {

      $index = zeroData($index, $count);

    }

    $count++;

    my $id     = $index->[ID];
    my $mask   = $index->[MASK];
    my $gpflag = $index->[GPFLAG];
    my $flag   = \$index->[FLAG];
    my $ratio  = $index->[RATIO];

    #print "[$gpflag]\n";

    # Set all masked genes to have flags equal to 'A'.


    if ($mask eq 'empty') {     # Empty spots.

      ${$flag} = 'A';

      $empty_spots++;
      next MASK;

    } elsif ($mask eq 'A') {    # PCR failures or bad spots.

      ${$flag} = 'A';

      $masked_spots++;
      next MASK;

    } elsif ($mask eq 'M1') {

      ${$flag} = 'A';

      $masked_spots++;
      next MASK;

    } elsif (not $mask) {

      ${$flag} = 'A';

      $masked_spots++;
      next MASK;

    } elsif ($mask eq 'T') {    # Tags or markers.

      ${$flag} = 'P';

    } elsif ($mask eq 'Y') {    # Budding yeast genes.

      ${$flag} = 'P';

    } elsif ($mask eq 'Q') {    # Bacterial control spots.

      ${$flag} = 'P';

    }

    # Set all 'negative'/'empty' genes to have flags equal to 'A'.

    if ($id eq 'empty') {

      ${$flag} = 'A';

      $empty_spots++;
      next MASK;

    } elsif ($id eq 'negative') {

      ${$flag} = 'A';

      $negative_spots++;
      next MASK;

    }

    # Set all 'not found'/'error' genes to have flags equal to 'A'.

    if ($gpflag >= 0) {

      ${$flag} = 'P';

    } elsif ($gpflag == -50) {

      ${$flag} = 'A';

      $not_found_spots++;
      next MASK;

    } elsif ($gpflag == -75) {

      ${$flag} = 'A';

      $bad_spots++;
      next MASK;

    } elsif ($gpflag == -100) {

      ${$flag} = 'A';

      $bad_spots++;
      next MASK;

    }

    if ($ratio == 0) {

      ${$flag} = 'A';

      $zero_ratio_spots++;
      next MASK;

    }

    # Set all flags of marginal genes equal to 'M'.

    if ( ($mask eq 'M') and ($gpflag >= 0) ) {

      ${$flag} = 'M';

      $marginal_spots++;
      next MASK;

    }

  }

  print $fh
    "Masked spot information:\n\n" .
    "Number of masked spots:\t\t\t" . $masked_spots . ".\n" .
    "Number of not found spots:\t\t" . $not_found_spots . ".\n" .
    "Number of empty spots:\t\t\t" . $empty_spots . ".\n" .
    "Number of negative spots:\t\t" . $negative_spots . ".\n" .
    "Number of spots flagged bad by user:\t" . $bad_spots . ".\n" .
    "Number of spots with zero ratios:\t" . $zero_ratio_spots . ".\n" .
    "Number of marginal spots:\t\t" . $marginal_spots . ".\n\n";

  return TRUE;

} # maskData


#############################################################################
#  bgCutOff
#  --------

sub bgCutOff {

  my $fh = shift;

  # Set flags to 'A' for genes that do not have pixel intensities
  #   above background.

  # Also set flags to 'A' for genes with very low or negative signals;
  #   commented out at the moment.

  my $count = 0;

  my $upper_bg_cut = $gOpts{upperbg};
  my $lower_bg_cut = $gOpts{lowerbg};

 BG:
  foreach my $index (@{$gDataArray}) {

    my $flag = \$index->[FLAG];
    my $mask = $index->[MASK];

    my $b635sd = $index->[B635SD];
    my $b532sd = $index->[B532SD];

    my $sig_expt = $index->[SIG_EXPT];
    my $sig_cont = $index->[SIG_CONT];

    next BG
      if $mask eq 'A';

    if ($b635sd < $upper_bg_cut and $b532sd < $upper_bg_cut) {

      if ($b635sd < $lower_bg_cut or $b532sd < $lower_bg_cut) {

        #my $id = $index->[ID];
        #print STDERR "[$id]\n";

        ${$flag} = 'A';
        $count++;

        next BG;

      }

    }

    #if ($sig_expt <= $sig_cut or $sig_cont <= $sig_cut) {

    #  ${$flag} = 'A';
    #  next BG;

    #}

  }

  print $fh
    "Using upper background cut-off:\t\t" . $upper_bg_cut . ".\n" .
    "Using lower background cut-off:\t\t" . $lower_bg_cut . ".\n\n" .
    "Number of spots discarded:\t\t" . $count . ".\n\n";

  return TRUE;

} # bgCutOff


#############################################################################
#  diamCutOff
#  ----------

sub diamCutOff {

  my $fh = shift;

  # Set flags to 'A' for spots that have a diameter
  #   lower than a set cut-off.

  my $count = 0;

  my $diam_cut = $gOpts{diamcut};

 DIAM:
  foreach my $index (@{$gDataArray}) {

    my $flag = \$index->[FLAG];
    my $mask = $index->[MASK];

    my $diam = $index->[DIAM];

    next DIAM
      if $mask eq 'A';

    if ($diam < $diam_cut) {

      ${$flag} = 'A';
      $count++;

    }

  }

  print $fh
    "Using minimum diameter cut-off:\t\t" . $diam_cut . ".\n\n" .
    "Number of spots discarded:\t\t" . $count . ".\n\n";

  return TRUE;

} # diamCutOff


#############################################################################
#  makeGS
#  ------

sub makeGS {

 FLAGS:
  foreach my $index (@{$gDataArray}) {

    my $flag   = $index->[FLAG];
    my $gsflag = \$index->[GSFLAG];

    if ($flag eq 'P') {
      ${$gsflag} = 'P';

    } elsif ($flag eq 'M') {
      ${$gsflag} = 'M';

    } elsif ($flag eq 'A') {
      ${$gsflag} = 'A';

    } else {
      #print "[$flag]\n";
      ${$gsflag} = '-';

    }

  }

  return TRUE;

} # makeGS


#############################################################################
#  filterData
#  ----------

sub filterData {

  my $file = shift;
  my $fh   = shift;

  if ($gOpts{bgcut} == TRUE) {

    print $fh "Using background SD cut-offs.\n\n";

    bgCutOff($fh)
      or Die("Cannot use background SD cut-offs in file $file.");

  } else {

    print $fh "Not using background SD cut-offs.\n\n";

  }

  # Change flags for spots with diameters less than cut-off.

  unless ($gOpts{diamcut} == 0) {   # This is an integer - 0 means false.

    print $fh "Using spot diameter cut-off.\n\n";

    diamCutOff($fh)
      or Die("Cannot use diameter cut-off in file $file.");

  } else {

    print $fh "Not using spot diameter cut-off.\n\n";

  }

  # Change flags to GeneSpring format.

  print STDERR "Changing flags to GeneSpring format.\n\n";

  makeGS()
    or Die("Cannot change flags to GeneSpring format in file $file.");

  return TRUE;

} # filterData


#############################################################################
#  calcSigRatio
#  ------------

sub calcSigRatio {

 CALC_SIG_RATIO:
  foreach my $index (@{$gDataArray}) {

    my $flag = $index->[FLAG];
    my $mask = $index->[MASK];

    next CALC_SIG_RATIO
      if $flag eq 'A';

    my $sig_ratio = 0;

    if ($index->[SIG_CONT] > 0) {
      $sig_ratio = $index->[SIG_EXPT]/$index->[SIG_CONT];
    }

    #$sig_ratio = 0
    #  if $sig_ratio < 0;

    $index->[SIG_RATIO] = $sig_ratio;

  }

  return TRUE;

} # calcSigRatio


#############################################################################
#  getFilteredData
#  ---------------

sub getFilteredData {

 GET_FILTER:
  foreach my $index (@{$gDataArray}) {

    $index->[NORM_RATIO1]    = $index->[RATIO];
    $index->[NORM_SIG_EXPT1] = $index->[SIG_EXPT];

  }

  return TRUE;

} # getFilteredData


#############################################################################
#  expandWindow
#  ------------

sub expandWindow {

  my $left   = shift;
  my $right  = shift;
  my $top    = shift;
  my $bottom = shift;

  my $array_col = shift;
  my $array_row = shift;

  my $maxwin_size = $gOpts{maxwin};

  if ($left == 1 and $right - $array_col <= $maxwin_size) {
    $right++;
  }

  if ($top == 1 and $bottom - $array_row <= $maxwin_size) {
    $bottom++;
  }

  if ($right == $gBigX and $array_col - $left <= $maxwin_size) {
    $left--;
  }

  if ($bottom == $gBigY and $array_row - $top <= $maxwin_size) {
    $top--;
  }

  return ($left, $right, $top, $bottom);

} # expandWindow


#############################################################################
#  getCount
#  --------

sub getCount {

  my $left   = shift;
  my $right  = shift;
  my $top    = shift;
  my $bottom = shift;

  my $norm_local = shift;

  my $count = 0;

 FORI:
  for (my $i = $left; $i <= $right; $i++) {

  FORJ:
    for (my $j = $top; $j <= $bottom; $j++) {

      my $pos = ($j - 1) * $gBigX + $i - 1;

      my $index = $gDataArray->[$pos];

      my $flag = $index->[FLAG];
      my $mask = $index->[MASK];

      next FORJ
        unless $flag eq 'P';

      if ($norm_local == NON_SPIKES) {

        next FORJ
          if $mask eq 'Q';

      } elsif ($norm_local == SPIKES) {

        next FORJ
          unless $mask eq 'Q';

      } elsif ($norm_local == ALL_SPOTS) {

        # Do nothing!

      } elsif ($norm_local == NON_CONTROLS) {

        next FORJ
          if $mask eq 'Q' or $mask eq 'T' or $mask eq 'Y';

      }

      $count++;

    }

  }

  return $count;

} # getCount


#############################################################################
#  increaseSquare
#  --------------

sub increaseSquare {

  my $left   = shift;
  my $right  = shift;
  my $top    = shift;
  my $bottom = shift;

  my $array_col  = shift;
  my $array_row  = shift;
  my $norm_local = shift;

  my $block_size = undef;

  if ($norm_local == SPIKES) {

    $block_size = $gOpts{spblocksize};

  } else {

    $block_size = $gOpts{blocksize};

  }

  my $count     = 0;
  my $old_count = 0;
  my $max_flag  = FALSE;

 INCREASE:
  while (1) {

    ($left, $right, $top, $bottom) =
      expandWindow($left, $right, $top, $bottom, $array_col, $array_row);

    $count =
      getCount($left, $right, $top, $bottom, $norm_local);

    #print "[$count] [$old_count] [$block_size]\n";

    last INCREASE
      if $count > $block_size;

    $max_flag = TRUE, last INCREASE
      if $count == $old_count;

    $old_count = $count;

  }

  return ($left, $right, $top, $bottom, $max_flag);

} # increaseSquare


#############################################################################
#  getSquareValues
#  ---------------

sub getSquareValues {

  my $curr_round = shift;

  my $left   = shift;
  my $right  = shift;
  my $top    = shift;
  my $bottom = shift;

  my $array_col  = shift;
  my $array_row  = shift;
  my $norm_local = shift;

  my $ratios = undef;

  my $total_sig_expt = 0;
  my $total_sig_cont = 0;

  # Find positions of spots in a square around a spot (+ and - window);
  #   gets ratio for each spot, and saves them into a list.

  my $count = 0;

 FORX:
  for (my $i = $left; $i <= $right; $i++) {

  FORY:
    for (my $j = $top; $j <= $bottom; $j++) {

      my $pos = ($j - 1) * $gBigX + $i - 1;

      my $index = $gDataArray->[$pos];

      my $flag = $index->[FLAG];
      my $mask = $index->[MASK];

      #print "[$flag] [$mask]\n";

      next FORY
        unless $flag eq 'P';

      if ($norm_local == NON_SPIKES) {

        next FORY
          if $mask eq 'Q';

      } elsif ($norm_local == SPIKES) {

        next FORY
          unless $mask eq 'Q';

      } elsif ($norm_local == ALL_SPOTS) {

        # Do nothing!

      } elsif ($norm_local == NON_CONTROLS) {

        next FORY
          if $mask eq 'Q' or $mask eq 'T' or $mask eq 'Y';

      }

      my $index_ratio = $index->[RATIO];

      my $sig_expt = $index->[SIG_EXPT];

      if ($gOpts{round1} == ALL_NON_CONT and
          $gOpts{round2} != NONE and
          $curr_round == 1) {

        $index->[COUNT_USED_IN_0]++;

      } elsif ($curr_round == 1) {

        $index->[COUNT_USED_IN_1]++;

      } elsif ($curr_round == 2) {

        my $use_round1_flag = $index->[FLAG_NORM_IN_0];

        if ($use_round1_flag == TRUE) {

          $index_ratio = $index->[NORM_RATIO0];

          $sig_expt = $index->[NORM_SIG_EXPT0];

          $index->[COUNT_USED_IN_1]++;

        }

      }

      push @{$ratios}, $index_ratio;

      $total_sig_expt += $sig_expt;
      $total_sig_cont += $index->[SIG_CONT];

      $count++;

    }

    #print "\n";

  }

  my $block_size = undef;

  if ($norm_local == SPIKES) {

    $block_size = $gOpts{spblocksize};

  } else {

    $block_size = $gOpts{blocksize};

  }

  my $max_flag = FALSE;

  # $count is number of present spots in the square.

  # If less than this,
  #  the square is expanded until a maximum window size is reached.

  unless ($count > $block_size) {

    ($left, $right, $top, $bottom, $max_flag) =
      increaseSquare($left, $right, $top, $bottom,
                     $array_col, $array_row, $norm_local);

    if ($max_flag == FALSE) {

      # This code is done if the maximum window size has not been reached,
      #   and the square has increased in size.

      $ratios = undef;

      $total_sig_expt = 0;
      $total_sig_cont = 0;

      # Note that recursion is being used here.

      ($ratios, $total_sig_expt, $total_sig_cont) =
        getSquareValues($curr_round, $left, $right, $top, $bottom,
                        $array_col, $array_row, $norm_local);

    }

  }

  return ($ratios, $total_sig_expt, $total_sig_cont);

} # getSquareValues


#############################################################################
#  checkSquareSize
#  ---------------

sub checkSquareSize {

  my $left   = shift;
  my $right  = shift;
  my $top    = shift;
  my $bottom = shift;

  my $array_col  = shift;
  my $array_row  = shift;
  my $norm_local = shift;

  # Increase block size by defined amount to account
  #   for absent spots in the square.

  # This is a heuristic to speed up the search!

  my $initial_block_size = undef;

  if ($norm_local == SPIKES) {

    $initial_block_size = $gOpts{spblockst};

  } else {

    $initial_block_size = $gOpts{blockstart};

  }

  my $count     = 0;
  my $old_count = 0;

 COUNT:
  while (1) {

    # Doesn't check for present spots, this is done later.

    $count = ($bottom - $top + 1) * ($right - $left + 1);

    last COUNT
      if $count > $initial_block_size;

    last COUNT
      if $count == $old_count;

    $old_count = $count;

    ($left, $right, $top, $bottom) =
      expandWindow($left, $right, $top, $bottom, $array_col, $array_row);

  }

  return ($left, $right, $top, $bottom);

} # checkSquareSize


#############################################################################
#  getFinalSquareValues
#  --------------------

sub getFinalSquareValues {

  my $index = shift;

  my $curr_round = shift;
  my $norm_local = shift;

  my $array_col = $index->[ARRAY_COL];
  my $array_row = $index->[ARRAY_ROW];

  my $window_size = $gOpts{window};

  my $left = $array_col - $window_size;

  $left = 1
    if $left < 1;

  my $top = $array_row - $window_size;

  $top = 1
    if $top < 1;

  my $right = $array_col + $window_size;

  $right = $gBigX
    if $right > $gBigX;

  my $bottom = $array_row + $window_size;

  $bottom = $gBigY
    if $bottom > $gBigY;

  # Make a square that must have a minimum number of spots in it;
  #   expand the square in two directions if less than this minimum.

  # This does not check whether all spots are present,
  #   and may use a larger square than is necessary.

  ($left, $right, $top, $bottom) =
    checkSquareSize($left, $right, $top, $bottom,
                    $array_col, $array_row, $norm_local);

  my $ratios = undef;

  my $total_sig_expt = 0;
  my $total_sig_cont = 0;

  ($ratios, $total_sig_expt, $total_sig_cont) =
    getSquareValues($curr_round, $left, $right, $top, $bottom,
                    $array_col, $array_row, $norm_local);

  my $norm_sig_factor = 0;

  if ($total_sig_cont > 0) {
    $norm_sig_factor = $total_sig_expt/$total_sig_cont;
  }

  return ($norm_sig_factor, $ratios);

} # getFinalSquareValues


#############################################################################
#  calcMedian
#  ----------

sub calcMedian {

  my $ratios = shift;

  my $median = -1;

  my @ratios = sort { $a <=> $b } @{$ratios};

  if (@ratios%2) {

    $median = $ratios[@ratios/2];

  } else {

    if ($ratios[@ratios/2] > 0) {

      $median = exp( (log($ratios[@ratios/2 - 1]) +
                      log($ratios[@ratios/2]) )/2);

    } else {

      $median = $ratios[@ratios/2];

    }

  }

  return $median;

} # calcMedian


#############################################################################
#  localNormalise
#  --------------

sub localNormalise {

  my $fh = shift;

  my $curr_round       = shift;
  my $final_round_flag = shift;

  my $norm_spots = shift;
  my $norm_local = shift;

  my $block_size = undef;

  if ($norm_local == SPIKES) {

    $block_size = $gOpts{spblocksize};

  } else {

    $block_size = $gOpts{blocksize};

  }

  my $count = 0;

  undef my @low_spots;
  undef my @zero_med;
  undef my @zero_cont;

 LOCAL_NORM:
  foreach my $index (@{$gDataArray}) {

    my $flag = $index->[FLAG];
    my $mask = $index->[MASK];

    #my $id = $index->[ID];
    #print "[$id] [$mask] [$flag]\n";

    next LOCAL_NORM
      if $flag eq 'A';

    if ($norm_spots == NON_SPIKES) {

      next LOCAL_NORM
        if $mask eq 'Q';

    } elsif ($norm_spots == SPIKES) {

      next LOCAL_NORM
        unless $mask eq 'Q';

    } elsif ($norm_spots == ALL_SPOTS) {

      # Do nothing!

    } elsif ($norm_spots == NON_CONTROLS) {

      next LOCAL_NORM
        if $mask eq 'Q' or $mask eq 'T' or $mask eq 'Y';

    }

    $count++;

    print STDERR $count . "\r";

    # Find a suitable local square to get normalisation factors.

    # $ratios is an array of ratios for all spots in this square.

    # $norm_sig_factor is the ratio for the total signals in this square.

    my ($norm_sig_factor, $ratios) =
      getFinalSquareValues($index, $curr_round, $norm_local);

    my $median = -1;

    if (defined $ratios) {
      $median = calcMedian($ratios);
    }

    my $no_spots = $#{$ratios};

    my $sig_ratio   = $index->[SIG_RATIO];
    my $index_ratio = $index->[RATIO];
    my $sig_expt    = $index->[SIG_EXPT];

    if ($gOpts{round2} == GLOBAL or
        ($gOpts{round2} == LOCAL and $final_round_flag != TRUE) ) {

      $index->[FLAG_NORM_IN_0] = TRUE;

      $index->[NORM_MED_RATIO0] = $median;

      $index->[NO_SPOTS0] = $no_spots;

      if ($median > 0) {

        if ($index_ratio > 0) {

          $index->[NORM_RATIO0] =
            $index_ratio/$median;

        }

      }

      $index->[NORM_SIG_FACTOR0] = $norm_sig_factor;

      if ($norm_sig_factor > 0) {

        if ($sig_ratio > 0) {

          $index->[NORM_SIG_RATIO0] =
            $sig_ratio/$norm_sig_factor;

        }

        if ($sig_expt > 0) {

          $index->[NORM_SIG_EXPT0] =
            $sig_expt/$norm_sig_factor;

        }

      }

    } else {

      # If following flag has been set to FALSE, then use original values
      #   in second normalisation round.

      # Else if flag has been set to TRUE, then use values from first
      #   normalisation round in second normalisation round.

      my $use_round1_flag = $index->[FLAG_NORM_IN_0];

      if ($curr_round == 2 and $use_round1_flag == TRUE) {

        $sig_ratio   = $index->[NORM_SIG_RATIO0];
        $index_ratio = $index->[NORM_RATIO0];
        $sig_expt    = $index->[NORM_SIG_EXPT0];

      }

      $index->[FLAG_NORM_IN_1] = TRUE;

      $index->[NORM_MED_RATIO1] = $median;

      $index->[NO_SPOTS1] = $no_spots;

      if ($median > 0) {

        if ($index_ratio > 0) {

          $index->[NORM_RATIO1] =
            $index_ratio/$median;

        }

      }

      $index->[NORM_SIG_FACTOR1] = $norm_sig_factor;

      if ($norm_sig_factor > 0) {

        # Calculate final normalised signals ratio;
        #   (signals ratio)/(normalisation factor).

        if ($sig_ratio > 0) {

          $index->[NORM_SIG_RATIO1] =
            $sig_ratio/$norm_sig_factor;

        }

        if ($sig_expt > 0) {

          $index->[NORM_SIG_EXPT1] =
            $sig_expt/$norm_sig_factor;

        }

      }

    }

    #if ($count < 10) {

    #  print
    #    "\n[$curr_round]\n[USED: $sig_ratio]\t[$index->[SIG_RATIO]]\t" .
    #    "[$index->[NORM_SIG_RATIO0]]\t[$index->[NORM_SIG_RATIO1]]\n";
    #  print
    #    "[USED: $index_ratio]\t[$index->[RATIO]]\t" .
    #    "[$index->[NORM_RATIO0]]\t[$index->[NORM_RATIO1]]\n";
    #  print
    #    "[USED: $sig_expt]\t[$index->[SIG_EXPT]]\t" .
    #    "[$index->[NORM_SIG_EXPT0]]\t[$index->[NORM_SIG_EXPT1]]\n\n";

    #}

    if ($median == 0) {

      push @zero_med, $index;

    }

    if ($index->[SIG_CONT] == 0) {

      push @zero_cont, $index;

    }

    if ($no_spots < $block_size) {

      push @low_spots, $index;

    }

  }

  if (@zero_med) {

    print $fh
      "Spots with zero median values:\n\n" .
      "ID\tBlock\tColumn\tRow\n";

    foreach my $index (@zero_med) {

      print $fh
        $index->[ID] . "\t" .
        $index->[ARRAY_BLOCK] . "\t" .
        $index->[BLOCK_COL] . "\t" .
        $index->[BLOCK_ROW] . "\n";

    }

    print $fh "\n";

  }

  if (@zero_cont) {

    print $fh
      "Spots with zero control values " .
      "(i.e., signal of control channel) :\n\n" .
      "ID\tBlock\tColumn\tRow\n\n";

    foreach my $index (@zero_cont) {

      print $fh
        $index->[ID] . "\t" .
        $index->[ARRAY_BLOCK] . "\t" .
        $index->[BLOCK_COL] . "\t" .
        $index->[BLOCK_ROW] . "\n";

    }

    print $fh "\n";

  }

  my $low_spots = @low_spots;

  my $string =
    $low_spots . ' spots (out of a total of ' . $count .
    ') used less than ' . $block_size .
    " local spots for normalisation.\n\n";

  print $fh $string
    if $gOpts{log} == TRUE;

  print STDERR $string;

  return TRUE;

} # localNormalise


#############################################################################
#  bactCalcMeans
#  -------------

sub bactCalcMeans {

  my $fh = shift;

  my $bact_norm_type = shift;

  my $bact_no     = 0;
  my $total_ratio = 0;

  my $total_bact_expt = 0;
  my $total_bact_cont = 0;

 DATA:
  foreach my $index (@{$gDataArray}) {

    my $id   = $index->[ID];
    my $mask = $index->[MASK];
    my $flag = $index->[FLAG];

    #print STDERR "[$mask]\n";

    next DATA
      unless $mask eq 'Q';

    next DATA
      unless $flag eq 'P';

    $bact_no++;

    $total_bact_cont += $index->[SIG_CONT];

    if ($bact_norm_type == GLOBAL) {

      $total_ratio += $index->[RATIO];

      $total_bact_expt += $index->[SIG_EXPT];

    } elsif ($bact_norm_type == LOCAL) {

      $total_ratio += $index->[NORM_RATIO0];

      $total_bact_expt += $index->[NORM_SIG_EXPT0];

    }

  }

  my $bact_ratio_mean = $total_ratio/$bact_no;
  my $bact_sig_mean   = 0;

  if ($total_bact_cont > 0 and $total_bact_expt > 0) {

    $bact_sig_mean = $total_bact_expt/$total_bact_cont;

  }

  return ($bact_no, $bact_ratio_mean, $bact_sig_mean);

} # bactCalcMeans


#############################################################################
#  bactGetMeans
#  ------------

sub bactGetMeans {

  my $fh = shift;

  my $bact_norm_type = shift;

  my ($bact_no, $bact_ratio_mean, $bact_sig_mean) =
    bactCalcMeans($fh, $bact_norm_type);

  my $bact_min = DEF_BACT_MIN;

  if ($bact_no > $bact_min) {

    my $sbact_ratio_mean = sprintf "%.5g", $bact_ratio_mean;
    my $sbact_sig_mean   = sprintf "%.5g", $bact_sig_mean;

    my $string = '';
    my $pre_string = '';

    if ($bact_norm_type == GLOBAL) {

      $pre_string = 'Mean ';

    } elsif ($bact_norm_type == LOCAL) {

      $pre_string = 'Local mean ';

    }

    $string =
      $pre_string .
      'ratio of control RNAs (expt/control) is ' .
      $sbact_ratio_mean . ".\n" .
      $pre_string .
      'ratio of signals for control RNAs (expt/control) is ' .
      $sbact_sig_mean . ".\n" .
      'Calculated from ' . $bact_no . " spots.\n";

    print $fh $string
      if $gOpts{log} == TRUE;

    print STDERR $string;

  } else {

    $bact_ratio_mean = 0;
    $bact_sig_mean   = 0;

    my $string =
      "Ratio of control RNAs could not be calculated.\n" .
      'There are only ' . $bact_no . " bacterial spots.\n";

    print $fh $string
      if $gOpts{log} == TRUE;

    print STDERR $string;

  }

  print $fh "\n"
    if $gOpts{log} == TRUE;

  print STDERR "\n";

  return ($bact_ratio_mean, $bact_sig_mean);

} # bactGetMeans


#############################################################################
#  bactNormalise
#  -------------

sub bactNormalise {

  my $bact_norm_type  = shift;
  my $bact_ratio_mean = shift;
  my $bact_sig_mean   = shift;

 BACT_NORM:
  foreach my $index (@{$gDataArray}) {

    my $flag = $index->[FLAG];
    my $mask = $index->[MASK];

    next BACT_NORM
      if $flag eq 'A';

    next BACT_NORM
      if $mask eq 'Q';

    $index->[NORM_MED_RATIO1]  = $bact_ratio_mean;
    $index->[NORM_SIG_FACTOR1] = $bact_sig_mean;

    if ($bact_ratio_mean) {

      if ($bact_norm_type == LOCAL) {

        if ($index->[NORM_RATIO0] > 0) {

          $index->[NORM_RATIO1] =
            $index->[NORM_RATIO0]/$bact_ratio_mean;

        }

      } elsif ($bact_norm_type == GLOBAL) {

        if ($index->[RATIO] > 0) {

          $index->[NORM_RATIO1] =
            $index->[RATIO]/$bact_ratio_mean;

        }

      }

    } else {

      $index->[NORM_RATIO1] = 0;

    }

    if ($bact_sig_mean) {

      if ($bact_norm_type == LOCAL) {

        if ($index->[NORM_SIG_RATIO0] > 0) {

          $index->[NORM_SIG_RATIO1] =
            $index->[NORM_SIG_RATIO0]/$bact_sig_mean;

        }

        if ($index->[NORM_SIG_EXPT0] > 0) {

          $index->[NORM_SIG_EXPT1] =
            $index->[NORM_SIG_EXPT0]/$bact_sig_mean;

        }

      } elsif ($bact_norm_type == GLOBAL) {

        if ($index->[SIG_RATIO] > 0) {

          $index->[NORM_SIG_RATIO1] =
            $index->[SIG_RATIO]/$bact_sig_mean;

        }

        if ($index->[SIG_EXPT] > 0) {

          $index->[NORM_SIG_EXPT1] =
            $index->[SIG_EXPT]/$bact_sig_mean;

        }

      }

    } else {

      $index->[NORM_SIG_RATIO1] = 0;

    }

  }

  return TRUE;

} # bactNormalise


#############################################################################
#  normaliseData
#  -------------

sub normaliseData {

  my $file = shift;
  my $fh   = shift;

  my $no_local_rounds = 0;

  my $r1_spots = NONE;
  my $r1_local = NONE;

  my $r2_spots = NONE;
  my $r2_local = NONE;

  if ($gOpts{round1} == NONE) {

    $r1_spots = NOT_USING;
    $r1_local = NOT_USING;

    if ($gOpts{round2} == LOCAL) {

      $no_local_rounds = 1;

      $r2_spots = NON_SPIKES;
      $r2_local = SPIKES;

    } elsif ($gOpts{round2} == GLOBAL or
             $gOpts{round2} == NONE or
             $gOpts{round2} == NOT_USING) {

      $no_local_rounds = 0;

      $r2_spots = NOT_USING;
      $r2_local = NOT_USING;

    }

  } elsif ($gOpts{round1} == ALL_NON_CONT) {

    $r1_spots = ALL_SPOTS;
    $r1_local = NON_CONTROLS;

    if ($gOpts{round2} == LOCAL) {

      $no_local_rounds = 2;

      $r2_spots = NON_SPIKES;
      $r2_local = SPIKES;

    } elsif ($gOpts{round2} == GLOBAL or
             $gOpts{round2} == NONE or
             $gOpts{round2} == NOT_USING) {

      $no_local_rounds = 1;

      $r2_spots = NOT_USING;
      $r2_local = NOT_USING;

    }

  }

  if ($no_local_rounds > 0) {

    print $fh
      'Doing ', $no_local_rounds,
      ' round', $no_local_rounds == 1 ? '' : 's',
      ' of local normalisation', $no_local_rounds == 1 ? '' : 's', ".\n\n"
        if $gOpts{log} == TRUE;

    print STDERR
      'Doing ', $no_local_rounds,
      ' round', $no_local_rounds == 1 ? '' : 's',
      ' of local normalisation', $no_local_rounds == 1 ? '' : 's', ".\n\n";

    my %print_spot_opt  = (NON_SPIKES,   'all spots except spikes',
                           SPIKES,       'spikes',
                           ALL_SPOTS,    'all spots',
                           NON_CONTROLS, 'all spots except controls');

    my %print_local_opt = (NON_SPIKES,   'all local spots except spikes',
                           SPIKES,       'local spikes',
                           ALL_SPOTS,    'all local spots',
                           NON_CONTROLS, 'all local spots except ' .
                                         'controls');

    #  Multiple rounds of normalisation can be done with the code.

    my $final_round_flag = FALSE;

    for (my $i = 1; $i <= $no_local_rounds; $i++) {

      my $curr_round = $i;

      if ($curr_round == $no_local_rounds) {
        $final_round_flag = TRUE;
      }

      my $string = '';

      my $norm_spots = NONE;
      my $norm_local = NONE;

      if ($gOpts{round1} == NONE or $curr_round == 2) {

        $norm_spots = $r2_spots;
        $norm_local = $r2_local;

      } elsif ($curr_round == 1) {

        $norm_spots = $r1_spots;
        $norm_local = $r1_local;

      }

      $string =
        "Normalising '" . $print_spot_opt{$norm_spots} .
        "' with '" . $print_local_opt{$norm_local} . "'.\n";

      my $block_size = undef;
      my $initial_block_size = undef;

      if ($norm_local == SPIKES) {

        $block_size = $gOpts{spblocksize};
        $initial_block_size = $gOpts{spblockst};

      } else {

        $block_size = $gOpts{blocksize};
        $initial_block_size = $gOpts{blockstart};

      }

      my $window = $gOpts{window};
      my $maxwin = $gOpts{maxwin};

      $string =
        "Local normalisation parameters for round:\t" . $curr_round .
        ".\n\n" . $string .
        "Using normalisation window size of:\t\t" . $window . ".\n" .
        "Using maximum normalisation window of:\t\t" . $maxwin . ".\n" .
        "Using normalisation block size of:\t\t" . $block_size . ".\n" .
        "Using initial block size of:\t\t\t" . $initial_block_size .
        ".\n\n";

      print $fh $string
        if $gOpts{log} == TRUE;

      print STDERR $string;

      # This is the main local normalisation sub-routine.

      localNormalise($fh, $curr_round, $final_round_flag, $norm_spots, $norm_local)
        or Die("Cannot normalise data in file $file.");

    }

  } else {

    my $string = "Not doing local normalisation.\n\n";

    print $fh $string
      if $gOpts{log} == TRUE;

    print STDERR $string;

  }

  if ($gOpts{round2} == GLOBAL) {

    my $bact_norm_type = GLOBAL;

    print STDERR
      "Calculating global mean ratio for bacterial controls.\n\n";

    my ($bact_ratio_mean, $bact_sig_mean) =
      bactGetMeans($fh, $bact_norm_type);

    # The next lines mean that the spikes have been locally adjusted in
    #   the first round before being globally applied to the other spots.

    if ($gOpts{round1} == ALL_NON_CONT) {

      $bact_norm_type = LOCAL;

      print STDERR
        "Calculating local mean ratio for bacterial controls.\n\n";

      ($bact_ratio_mean, $bact_sig_mean) =
        bactGetMeans($fh, $bact_norm_type);

      my $string =
        'Doing global normalisation/correction with mean of local ' .
        "bacterial ratios.\n\n";

      print $fh $string
        if $gOpts{log} == TRUE;

      print STDERR $string;

      bactNormalise($bact_norm_type, $bact_ratio_mean, $bact_sig_mean)
        or Die('Cannot calculate bacterial corrected ' .
                 "ratios in file $file.");

    } else {

      $bact_norm_type = GLOBAL;

      my $string =
        'Doing global normalisation/correction with mean of overall ' .
        "bacterial ratios.\n\n";

      print $fh $string
        if $gOpts{log} == TRUE;

      print STDERR $string;

      bactNormalise($bact_norm_type, $bact_ratio_mean, $bact_sig_mean)
        or Die('Cannot calculate bacterial corrected ' .
               "ratios in file $file.");

    }

  } elsif ($gOpts{round2} == NONE or $gOpts{round2} == NOT_USING) {

    my $string = "Not using bacterial controls.\n\n";

    print $fh $string
      if $gOpts{log} == TRUE;

    print STDERR $string;

  } else {

    # Do nothing!

  }

  return TRUE;

} # normaliseData


#############################################################################
#  initialiseGeneHash
#  ------------------

sub initialiseGeneHash {

  $gGeneHash = undef;

  foreach my $gene (@gGeneList) {

    $gGeneHash->{$gene}[NO_VAL]    = 0;     # No. of values for this gene.
    $gGeneHash->{$gene}[MARG_FLAG] = 0;     # Flag if marginal gene.
    $gGeneHash->{$gene}[MEAN]      = 0;     # Mean value for this gene.
    $gGeneHash->{$gene}[CV]        = 0;     # % CV for this gene.

  }

  return TRUE;

} # initialiseGeneHash


#############################################################################
#  checkPresents
#  -------------

sub checkPresents {

 FOR1:
  foreach my $index (@{$gDataArray}) {

    my $id   = $index->[ID];
    my $flag = $index->[FLAG];
    my $mask = $index->[MASK];

    #next FOR1
    #  if $mask eq 'T' or $mask eq 'Y' or $mask eq 'Q';

    my $lcid = lc $id;

    unless (exists $gGeneHash->{$lcid}) {

      delete $gGeneHash->{$lcid};
      next FOR1;

    }

    next FOR1
      unless $flag eq 'P';

    $gGeneHash->{$lcid}[NO_VAL]++;

    my $no_val = $gGeneHash->{$lcid}[NO_VAL];

    if (defined $index->[NORM_RATIO1] and
        $index->[NORM_RATIO1] > 0) {

      $gGeneHash->{$lcid}[$no_val+DATA_START-1] =
        log $index->[NORM_RATIO1];

    } else {

      $gGeneHash->{$lcid}[$no_val+DATA_START-1] = 0;

    }

  }

  return TRUE;

} # checkPresents


#############################################################################
#  checkMarginals
#  --------------

sub checkMarginals {

  my $fh = shift;

  print $fh "Genes with only marginal spots:\n\n";

  my $marg_count = 0;

 FOR2:
  foreach my $index (@{$gDataArray}) {

    my $id   = $index->[ID];
    my $flag = $index->[FLAG];
    my $mask = $index->[MASK];

    #next FOR2
    #  if $mask eq 'T' or $mask eq 'Y' or $mask eq 'Q';

    my $lcid = lc $id;

    unless (exists $gGeneHash->{$lcid}) {

      delete $gGeneHash->{$lcid};
      next FOR2;

    }

    next FOR2
      unless $flag eq 'M';

    if ($gGeneHash->{$lcid}[NO_VAL] == 0 or
        $gGeneHash->{$lcid}[MARG_FLAG] > 0) {

      $marg_count++, print $fh $id . "\n"
        if $gGeneHash->{$lcid}[MARG_FLAG] == 1;

      $gGeneHash->{$lcid}[NO_VAL]++;
      $gGeneHash->{$lcid}[MARG_FLAG]++;

      my $no_val = $gGeneHash->{$lcid}[NO_VAL];

      if (defined $index->[NORM_RATIO1] and $index->[NORM_RATIO1] > 0) {

        $gGeneHash->{$lcid}[$no_val+DATA_START-1] =
          log $index->[NORM_RATIO1];

      } else {

        $gGeneHash->{$lcid}[$no_val+DATA_START-1] = 0;

      }

    }

  }

  if ($marg_count > 0) {

    print $fh
      "\nThere are " . $marg_count . " marginal genes.\n\n";

  } else {

    print $fh
      "There are no marginal genes.\n\n";

  }

  return TRUE;

} # checkMarginals


#############################################################################
#  calcGeneMean
#  ------------

sub calcGeneMean {

 FOR3:
  foreach my $id (keys %{$gGeneHash}) {

    my $no_val = $gGeneHash->{$id}[NO_VAL];

    next FOR3
      if $no_val == 0;

    my $total = 0;

    for (my $i = DATA_START; $i <= ($no_val + DATA_START - 1); $i++) {

      $total += $gGeneHash->{$id}[$i];

    }

    my $mean = $total/$no_val;

    $mean = exp $mean;

    $gGeneHash->{$id}[MEAN] = $mean;

  }

  return TRUE;

} # calcGeneMean


#############################################################################
#  calcGeneCV
#  ----------

sub calcGeneCV {

 FOR4:
  foreach my $id (keys %{$gGeneHash}) {

    my $no_val = $gGeneHash->{$id}[NO_VAL];
    my $mean   = $gGeneHash->{$id}[MEAN];

    next FOR4
      if $no_val == 0;

    if ($no_val == 1) {

      $gGeneHash->{$id}[CV] = 0;   # CV not calculated if only one value.
      next FOR4;

    }

    my $ssd = 0; # Sum of squared differences between values and the mean.

    for (my $i = DATA_START; $i <= ($no_val + DATA_START - 1); $i++) {

      $ssd += (exp($gGeneHash->{$id}[$i]) - $mean)**2;

    }

    my $std_dev = sqrt($ssd/($no_val - 1) );

    my $cv = 100.0 * ($std_dev/$mean);

    $gGeneHash->{$id}[CV] = $cv;

  }

  return TRUE;

} # calcGeneCV


#############################################################################
#  calcTotals
#  ----------

sub calcTotals {

  my $fh = shift;

  my $total_cont = 0;
  my $total_expt = 0;

  my $total_norm_expt0 = 0;
  my $total_norm_expt1 = 0;

 FOR5:
  foreach my $index (@{$gDataArray}) {

    my $flag = $index->[FLAG];
    my $mask = $index->[MASK];

    next FOR5
      unless $flag eq 'P';

    next FOR5
      if $mask eq 'Q' or $mask eq 'T' or $mask eq 'Y';

    $total_cont += $index->[SIG_CONT];
    $total_expt += $index->[SIG_EXPT];

    $total_norm_expt0 += $index->[NORM_SIG_EXPT0];
    $total_norm_expt1 += $index->[NORM_SIG_EXPT1];

  }

  my $stotal_cont = sprintf "%.1f", $total_cont;
  my $stotal_expt = sprintf "%.1f", $total_expt;

  my $stotal_nexpt0 = sprintf "%.1f", $total_norm_expt0;
  my $stotal_nexpt1 = sprintf "%.1f", $total_norm_expt1;

  my $stotal_ratio = sprintf "%.6g", $total_expt/$total_cont;

  my $string =
    "Overall array signals:\n\n" .
    "Total signal control:\t\t\t\t" . $stotal_cont . ".\n" .
    "Total signal experiment:\t\t\t" . $stotal_expt . ".\n" .
    "Ratio of total signals:\t\t\t\t" . $stotal_ratio . ".\n\n" .
    "Total intermed. norm. signal experiment:\t" . $stotal_nexpt0 . ".\n" .
    "Total final norm. signal experiment:\t\t" . $stotal_nexpt1 . ".\n\n";

  print $fh $string
    if $gOpts{log} == TRUE;

  print STDERR $string;

  return TRUE;

} # calcTotals


#############################################################################
#  calcHighCVs
#  -----------

sub calcHighCVs {

  my $fh = shift;

  my $high_cvs = shift;

 FOR6:
  foreach my $id (sort { lc $a cmp lc $b } @{$high_cvs}) {

    my $no_val = $gGeneHash->{$id}[NO_VAL];
    my $cv     = $gGeneHash->{$id}[CV];

    undef my @list;

    for (my $i = DATA_START; $i <= ($no_val + DATA_START - 1); $i++) {

      push @list, $gGeneHash->{$id}[$i];

    }

    @list = sort { $a <=> $b } @list;

    my $max_ratio = exp($list[$no_val-1])/exp($list[0]);

    printf $fh "%s\t%d\t%.5g\t%.5g\n", $id, $no_val, $cv, $max_ratio;

  }

  print $fh "\n";

  return TRUE;

} # calcHighCVs


#############################################################################
#  calcQCforReplicates
#  -------------------

sub calcQCforReplicates {

  my $fh = shift;

  my $zero  = 0;
  my $once  = 0;
  my $twice = 0;
  my $more  = 0;

  my $high_cvs = undef;

  my $count = 0;
  my $total = 0;

 FOR7:
  foreach my $id (keys %{$gGeneHash}) {

    my $no_val = $gGeneHash->{$id}[NO_VAL];
    my $cv     = $gGeneHash->{$id}[CV];

    #print "[$id]\n";
    #  if $no_val == 0

    SWITCH: {

      $zero++,  last SWITCH
        if $no_val == 0;

      $once++,  last SWITCH
        if $no_val == 1;

      $twice++, last SWITCH
        if $no_val == 2;

      $more++,  last SWITCH
        if $no_val >= 3;

    }

    # Make list of genes with high CVs.

    push @{$high_cvs}, $id
      if $cv >= DEF_CV_CUT;

    # Used to calculate mean CV.

    if ($no_val >= 1) {

      $count++;
      $total += $cv;

    }

  }

  my $mean_cv = 0;

  if ($count > 0) {
    $mean_cv = sprintf "%.5g", $total/$count;
  } else {
    $mean_cv = 0;
  }

  print $fh
    "Replicate spot information:\n\n" .

    "Zero spots:\t\t" . $zero  . ".\n" .
    "One spot:\t\t"   . $once  . ".\n" .
    "Two spots:\t\t"  . $twice . ".\n" .

    "More than two spots:\t" . $more . ".\n\n" .

    "Average % CV:\t\t" . $mean_cv . ".\n\n" .

    'Genes with CVs higher than ' .
    DEF_CV_CUT . "%:\n\n";

  my $file = $gFileParts->{file};

  if (defined $high_cvs) {

    print $fh
      "Gene\tSpots\t% CV\tMax. ratio\n\n";

    calcHighCVs($fh, $high_cvs)
      or Die("Cannot calculate genes with high CVs in $file.");

    my $no_high_cvs = $#{$high_cvs} + 1;

    print $fh
      "There are " . $no_high_cvs . " genes with high CVs.\n\n";

  } else {

    print $fh
      'No genes with CVs higher than ' . DEF_CV_CUT . "%.\n\n";

  }

  return TRUE;

} # calcQCforReplicates


#############################################################################
#  printNormGenes
#  --------------

sub printNormGenes {

  my $file_ng = $gFileParts->{disk} .
                $gFileParts->{dir}  . $gDirSep .
                $gFileParts->{base} . '_';

  my $extra_ul = '';

  $extra_ul = $gOpts{filetext} ? '_' : '';

  $file_ng .=  $gOpts{filetext} . $extra_ul . 'norm_genes.txt';

  open NG, '>' . $file_ng
    or Die("Cannot create file $file_ng: $!");

  print NG
    "Gene name\t" .
    "Mean of ratios";

  if ($gOpts{genestats}) {

    print NG
      "\t% CV of ratios\t" .
      "Number of spots\t" .
      "Ratio Values (spread over multiple columns)\n";

  } else {

    print NG "\n";

  }

 FOR8:
  foreach my $id (sort { lc $a cmp lc $b } keys %{$gGeneHash}) {

    my $no_val = $gGeneHash->{$id}[NO_VAL];
    my $mean   = $gGeneHash->{$id}[MEAN];
    my $cv     = $gGeneHash->{$id}[CV];

    printf NG "%s\t", $id;

    if ($mean > 0) {
      printf NG "%.3f", $mean;
    }

    unless ($gOpts{genestats}) {
      print NG "\n";
      next FOR8;
    } else {
      print NG "\t";
    }

    if ($cv > 0) {
      printf NG "%.3f", $cv;
    }
    print NG "\t";

    printf NG "%d", $no_val;

    if ($no_val > 0) {

      print NG "\t";

      for (my $i = DATA_START; $i <= ($no_val + DATA_START - 1); $i++) {

        printf NG "%.5g", exp($gGeneHash->{$id}[$i]);

        unless ($i == ($no_val + DATA_START - 1) ) {

          print NG "\t";

        }

      }

    }

    print NG "\n";

  }

  close NG
    or Die("Cannot close file $file_ng: $!");

  return TRUE;

} # printNormGenes


#############################################################################
#  spotsToGenes
#  ------------

sub spotsToGenes {

  my $fh = shift;

  my $file = $gFileParts->{file};

  # Initialise gene list hash.

  initialiseGeneHash()
    or Die("Cannot initialise global gene hash in file $file.");

  # Remove genes that are flagged as absent/marginal.

  checkPresents()
    or Die("Cannot check for present genes in $file.");

  # Check for genes that only have one marginal probe set.

  checkMarginals($fh)
    or Die("Cannot add genes that only have marginal probes in $file.");

  # Calculate mean for each gene.

  calcGeneMean()
    or Die("Cannot calculate means for genes in $file.");

  # Calculate standard deviation and
  #   coefficient of variance (CV) for each gene.

  calcGeneCV()
    or Die("Cannot calculate CVs for genes in $file.");

  # Calculate information about total signals.

  calcTotals($fh)
    or Die("Cannot calculate totals for signals of genes in $file.");

  # Calculate QC information about replicate spots.

  calcQCforReplicates($fh)
    or Die("Cannot calculate QC information for replicate spots in $file.");

  printNormGenes()
    or Die("Cannot print normalised ratios for genes in $file.");

  return TRUE;

} # spotsToGenes


#############################################################################
#  printNormSpots
#  --------------

sub printNormSpots {

  my $file_ns = $gFileParts->{disk} .
                $gFileParts->{dir}  . $gDirSep .
                $gFileParts->{base} . '_';

  my $extra_ul = '';

  $extra_ul = $gOpts{filetext} ? '_' : '';

  $file_ns .=  $gOpts{filetext} . $extra_ul . 'norm_spots.txt';

  open NS, '>' . $file_ns
    or Die("Cannot create file $file_ns: $!");

  print NS
    "Index\t" .
    "Array block\t" .
    "Block column\t" .
    "Block row\t" .
    "Array column\t" .
    "Array row\t" .
    "Ratio\t" .
    "Array X Pos\t" .
    "Array Y Pos\t" .
    "ID\t" .
    "Flag\t" .
    "Mask\t" .
    "Diameter\t" .
    "% > BG + 2SD for Cy5\t" .
    "% > BG + 2SD for Cy3\t" .
    "% saturated pixels for Cy5\t" .
    "% saturated pixels for Cy3\t" .
    "Median of ratios\t" .
    "Ratio of medians\t" .
    "Median experiment - BG\t" .
    "Median control - BG\t" .
    "Trust value\t" .
    "Average log intensity\t" .
    "Ratio (calculated from signals)\t" .

    "Intermediate number of spots used for normalisation\t" .
    "Intermediate normalisation factor (ratios)\t" .
    "Intermediate normalised ratio (ratios)\t" .
    "Intermediate normalisation factor (signals)\t" .
    "Intermediate normalised ratio (signals)\t" .
    "Intermediate normalised signal (experiment)\t" .
    "Intermediate normalisation flag\t" .
    "Number of times used during intermediate normalisation\t" .

    "Final number of spots used for normalisation\t" .
    "Final normalisation factor (ratios)\t" .
    "Final normalised ratio (ratios)\t" .
    "Final normalisation factor (signals)\t" .
    "Final normalised ratio (signals)\t" .
    "Final normalised signal (experiment)\t" .
    "Final normalisation flag\t" .
    "Number of times used during final normalisation\n";

 PRINT:
  foreach my $index (@{$gDataArray}) {

    printf NS "%d\t",   $index->[INDEX];

    unless ($index->[ARRAY_BLOCK]) {

      print NS "\t" x 29;

      print NS "\n";

      next PRINT;

    }

    printf NS "%d\t",   $index->[ARRAY_BLOCK];
    printf NS "%d\t",   $index->[BLOCK_COL];
    printf NS "%d\t",   $index->[BLOCK_ROW];
    printf NS "%d\t",   $index->[ARRAY_COL];
    printf NS "%d\t",   $index->[ARRAY_ROW];
    printf NS "%.5g\t", $index->[RATIO];
    printf NS "%d\t",   $index->[X_POS];
    printf NS "%d\t",   $index->[Y_POS];
    printf NS "%s\t",   $index->[ID];
    printf NS "%s\t",   $index->[FLAG];
    printf NS "%s\t",   $index->[MASK];
    printf NS "%d\t",   $index->[DIAM];
    printf NS "%d\t",   $index->[B635SD];
    printf NS "%d\t",   $index->[B532SD];
    printf NS "%d\t",   $index->[F635SAT];
    printf NS "%d\t",   $index->[F532SAT];
    printf NS "%.5g\t", $index->[MED_RATIO];
    printf NS "%.5g\t", $index->[RATIO_MED];
    printf NS "%d\t",   $index->[SIG_EXPT];
    printf NS "%d\t",   $index->[SIG_CONT];
    printf NS "%d\t",   $index->[TRUST];

    unless ($index->[AVG_LOG_SIG] == 0) {
      printf NS "%.5g", $index->[AVG_LOG_SIG];
    }
    print NS "\t";

    unless ($index->[SIG_RATIO] == 0) {
      printf NS "%.5g", $index->[SIG_RATIO];
    }
    print NS "\t";

    unless ($index->[NO_SPOTS0] == 0) {
      printf NS "%d",   $index->[NO_SPOTS0];
    }
    print NS "\t";

    unless ($index->[NORM_MED_RATIO0] == 0) {
      printf NS "%.5g", $index->[NORM_MED_RATIO0];
    }
    print NS "\t";

    unless ($index->[NORM_RATIO0] == 0) {
      printf NS "%.5g", $index->[NORM_RATIO0];
    }
    print NS "\t";

    unless ($index->[NORM_SIG_FACTOR0] == 0) {
      printf NS "%.5g", $index->[NORM_SIG_FACTOR0];
    }
    print NS "\t";

    unless ($index->[NORM_SIG_RATIO0] == 0) {
      printf NS "%.5g", $index->[NORM_SIG_RATIO0];
    }
    print NS "\t";

    unless ($index->[NORM_SIG_EXPT0] == 0) {
      printf NS "%.5g", $index->[NORM_SIG_EXPT0];
    }
    print NS "\t";

    printf NS "%d\t", $index->[FLAG_NORM_IN_0];
    printf NS "%d\t", $index->[COUNT_USED_IN_0];

    unless ($index->[NO_SPOTS1] == 0) {
      printf NS "%d",   $index->[NO_SPOTS1];
    }
    print NS "\t";

    unless ($index->[NORM_MED_RATIO1] == 0) {
      printf NS "%.5g", $index->[NORM_MED_RATIO1];
    }
    print NS "\t";

    unless ($index->[NORM_RATIO1] == 0) {
      printf NS "%.5g", $index->[NORM_RATIO1];
    }
    print NS "\t";

    unless ($index->[NORM_SIG_FACTOR1] == 0) {
      printf NS "%.5g", $index->[NORM_SIG_FACTOR1];
    }
    print NS "\t";

    unless ($index->[NORM_SIG_RATIO1] == 0) {
      printf NS "%.5g", $index->[NORM_SIG_RATIO1];
    }
    print NS "\t";

    unless ($index->[NORM_SIG_EXPT1] == 0) {
      printf NS "%.5g", $index->[NORM_SIG_EXPT1];
    }
    print NS "\t";

    printf NS "%d\t", $index->[FLAG_NORM_IN_1];
    printf NS "%d\n", $index->[COUNT_USED_IN_1];

  }

  close NS
    or Die("Cannot close file $file_ns: $!");

  return TRUE;

} # printNormSpots


#############################################################################
#  indexLine
#  ---------

sub indexLine {

  my $line = shift;

  my @array = split /\t/, $line;

  my $block = $array[0];

  my $block_col = $array[1];    # Column number within the block.
  my $block_row = $array[2];    # Row number within the block.

  # Calculate absolute column, row and index position
  #   (within the micro-array).

  my $array_col =
    ( ($block - 1)%$gBlocksArray) * $gColsBlock + $block_col;

  my $array_row =
    (int( ($block - 1)/$gBlocksArray) ) * $gRowsBlock + $block_row;

  # - 1 at the end of the next equation so that array starts at 0, not 1.

  my $index =
    ($array_row - 1) * $gBlocksArray * $gColsBlock + $array_col - 1;

  return $index;

} # indexLine


#############################################################################
#  printAllData
#  ------------

sub printAllData {

  my $fh = shift;

  my $file_ad = $gFileParts->{disk} .
                $gFileParts->{dir}  . $gDirSep .
                $gFileParts->{base} . '_';

  my $extra_ul = '';

  $extra_ul = $gOpts{filetext} ? '_' : '';

  $file_ad .=  $gOpts{filetext} . $extra_ul . 'all_data.txt';

  open AD, '>' . $file_ad
    or Die("Cannot create file $file_ad: $!");

  my $file = $gFileParts->{file};

  open FILE, $file
    or Die("Cannot open file $file: $!");

  my $header_flag = FALSE;

 LINE:
  while (defined (my $line = <FILE>) ) {

    $line = trimLine($line);

    if ($header_flag == FALSE) {

      # $header_flag is set to TRUE when at the end of the header.

      print AD $line;

      if ($line =~ /^Block\s+/) {

        $header_flag = TRUE;

        print AD
          "\tLocally Normalised Ratio\t" .
          "Integrated Flag\t" .
          "Ratio Normalised using Bacterial Controls\n";

      } elsif ($line =~ /^NormalizationMethod/) {

        my $script = (split /$gDirSep/, $0)[-1];

        print AD
          "\tLocal Normalisation Script: " . $script .
          ' (v ' . VERSION . ").\n";

      } else {

        print AD "\n";

      }

      next LINE;

    }

    unless ($header_flag == TRUE) {
      Die("Cannot read GenePix file $file correctly.");
    }

    my $index = indexLine($line);

    # The following line removes plate information which is useful
    #   for ArrayExpress.

    #$line =~ s/\d+\-[A-H]\d+_([^\t]+?)(_A?\d+)?(_[FR])?\t/$1\t/;

    print AD $line . "\t";

    if ($gOpts{round2} == NONE) {

      if ($gDataArray->[$index][NORM_RATIO1]) {
        printf AD "%.5g", $gDataArray->[$index][NORM_RATIO1];
      }

      printf AD "\t%s\t", $gDataArray->[$index][FLAG];

    } else {

      if ($gDataArray->[$index][NORM_RATIO0]) {
        printf AD "%.5g", $gDataArray->[$index][NORM_RATIO0];
      }

      printf AD "\t%s\t", $gDataArray->[$index][FLAG];

      if ($gDataArray->[$index][NORM_RATIO1]) {
        printf AD "%.5g", $gDataArray->[$index][NORM_RATIO1];
      }

    }

    print AD "\n";

  }

  close FILE
    or Die("Cannot close file $file: $!");

  close AD
    or Die("Cannot close file $file_ad: $!");

  return TRUE;

} # printAllData


#############################################################################
#  printData
#  ---------

sub printData {

  my $file = shift;
  my $fh   = shift;

  if ($gOpts{spotnorm} == TRUE) {

    print STDERR "Printing normalised data for good spots.\n\n";

    printNormSpots()
      or Die("Cannot print normalised data for file $file.");

  }

  # Only need a gene list if gene averaged data required.

  if ($gOpts{genenorm} == TRUE) {

    # Well known statement to make gene list unique.

    @gGeneList =
      sort { lc $a cmp lc $b }
      keys %{ { map { $_, 1 } @gGeneList} };

    print STDERR "Printing normalised data for good genes.\n\n";

    spotsToGenes($fh)
      or Die("Cannot convert spot data to gene data in file $file.");

  }

  if ($gOpts{alldata} == TRUE) {

    print STDERR "Printing all data for all spots on array.\n\n";

    printAllData($fh)
      or Die("Cannot print all data for all spots in file file.");

  }

  return TRUE;

} # printData


#############################################################################
#  handleFiles
#  -----------

sub handleFiles {

  my $no_local_rounds = shift;

  my @files = sort { $a cmp $b or $a <=> $b } glob $gOpts{file};

 FILES:
  foreach my $file (@files) {

    my $start_time = time();

    Warn("Cannot check file: $file."), next FILES
      unless checkTextFile($file);

    $gFileParts = fileParts($file);

    $file = $gFileParts->{file};

    # Many functions output information into the log file.

    my $fh = openLog($file);

    my $script = (split /$gDirSep/, $0)[-1];

    my $string =
      'Using script: ' . $script .
      ' (v ' . VERSION . ").\n\n";

    print $fh $string
      if $gOpts{log} == TRUE;

    print STDERR $string;

    my $time_date = getTimeDate();

    $string = "Analysing file: $file at " . $time_date . ".\n\n";

    print $fh $string
      if $gOpts{log} == TRUE;

    print STDERR $string;

    # Main block of code for data input.

    readData($file, $fh)
      or Die("Cannot read data in file $file.");


    # Identify genes that are PCR failures, removed by user, etc.

    print STDERR "Masking bad spots.\n\n";

    maskData($fh)
      or Die("Cannot mask data in file $file.");

    # Change flags for weaker spots.

    filterData($file, $fh)
      or Die("Cannot filter data in file $file.");

    # Calculate ratios of Cy5 and Cy3 channels.

    print STDERR "Calculating signal ratios.\n\n";

    calcSigRatio()
      or Die("Cannot calculate signal ratios in file $file.");

    # Main normalisation block of code below!

    if ($gOpts{round1} == NONE and $gOpts{round2} == NONE) {

      $string = "Not doing any normalisation steps.\n\n";

      print $fh $string
        if $gOpts{log} == TRUE;

      print STDERR $string;

      getFilteredData()
        or Die("Cannot get filtered un-normalised data in file $file.");

    } else {

      normaliseData($file, $fh)
        or Die("Cannot normalise data in file $file.");

    }

    #print Dumper $gDataArray;

    # Main block of code for data output.

    printData($file, $fh)
      or Die("Cannot print data in file $file.");

    #print Dumper $gGeneHash;

    $string = "Finished analysis for $file.\n\n";

    print $fh $string
      if $gOpts{log} == TRUE;

    print STDERR $string;

    my $end_time = time();

    my $diff_time = $end_time - $start_time;

    $string = "Time taken: " . $diff_time . " seconds.\n\n";

    print $fh $string
      if $gOpts{log} == TRUE;

    print STDERR $string;

    if ($gOpts{log} == TRUE) {

      close $fh
        or Die("Cannot close log for file $file: $!");

    }

    setGlobals()
      or Die('Cannot reset global variables.');

  }

  print STDERR "Finished all selected files.\n\n";

  return TRUE;

} # handleFiles


#############################################################################
#  startAnal
#  ---------

sub startAnal {

  unless (exists $gOpts{file} and defined $gOpts{file}) {

    my $text = 'No files selected.';

    if ($gOpts{tk} == TRUE) {

      showWarning($text)
        or Die($text);

      return FALSE;

    } else {

      $text .= "\nNeed to specify at least one file with --file option.\n";

      showUsage($text)
        or Die($text);

      return FALSE;

    }

  }

  setOpts()
    or Die('Cannot check options.');

  #print Dumper \%gOpts;

  # Only need a gene list if gene averaged data required.

  if ($gOpts{genenorm} == TRUE) {

    undef @gGeneList;

    # If gene list file is specified - read once at beginning.

    if ($gOpts{extgene}) {

      #print STDERR "Using gene list file $gOpts{genefile}.\n";

      readGeneFile()
        or Die("Cannot read gene list file $gOpts{genefile}.");

    }

  }

  handleFiles()
    or Die('Cannot process files.');

  resetTextOpts()
    or Die('Cannot reset text options.');

  return TRUE;

} # startAnal


#############################################################################
#  doCmd
#  -----

sub doCmd {

  my @opts = qw(help
                os=s
                file=s
                filetext=s
                genefile=s
                extgene=s
                window=s
                maxwin=s
                blocksize=s
                blockstart=s
                spblocksize=s
                spblockst=s
                round1=s
                round2=s
                bgcut=s
                upperbg=s
                lowerbg=s
                diamcut=s
                spotnorm=s
                genenorm=s
                genestats=s
                alldata=s
                log=s
                info=s);

  if (GetOptions(\%gOpts, @opts) == 0 or
      exists $gOpts{help}) {

    showUsage()
      or Die('Cannot show usage.');

    return FALSE;

  }

  startAnal()
    or Die('Cannot start analysis.');

  return TRUE;

} # doCmd


#############################################################################
#  openFiles
#  ---------

sub openFiles {

  require Tk::FileSelect;

  my %gui_opts = (-directory     => cwd(),
                  -filter        => '*.gpr',
                  -filelabel     => 'Or enter file name:',
                  -filelistlabel => 'Files',
                  -dirlabel      => 'Path',
                  -dirlistlabel  => 'Directories',
                  -width         => 30,
                  -verify        => ['-T'],     # Accept only text files.
                  -selectmode    => 'multiple');

  my $fs = $gTk->FileSelect(%gui_opts);

  my @files = $fs->Show;

  $gOpts{file} = join ' ', @files
    if defined $files[0];

  return TRUE;

} # openFiles


#############################################################################
#  openGeneList
#  ------------

sub openGeneList {

  require Tk::FileSelect;

  my %gui_opts = (-directory     => cwd(),
                  -filter        => '*.txt',
                  -filelabel     => 'Or enter file name:',
                  -filelistlabel => 'Files',
                  -dirlabel      => 'Path',
                  -dirlistlabel  => 'Directories',
                  -width         => 30,
                  -verify        => ['-T'],     # Accept only text files.
                  -selectmode    => 'single');

  my $fs = $gTk->FileSelect(%gui_opts);

  my @files = $fs->Show;

  $gOpts{genefile} = join ' ', @files
    if defined $files[0];

  return TRUE;

} # openGeneList


#############################################################################
#  fileFrame
#  ---------

sub fileFrame {

  my $top_frame = shift;

  my %gui_opts = (-text       => 'Files to be processed:',
                  -foreground => 'purple');

  my %pack_opts = (-padx => 2,
                   -pady => 2);

  my $file_label =
    $top_frame->Label(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-text             => 'File selector',
               -foreground       => 'dark green',
               -activebackground => 'dark green',
               -borderwidth      => 3,
               -relief           => 'raised',
               -command          => \&openFiles);

  my $open_dir_button =
    $top_frame->Button(%gui_opts)->pack();

  %gui_opts = (-text       => 'Append following text to output files:',
               -foreground => 'dark green');

  my $win_label =
    $top_frame->Label(%gui_opts)->pack(%pack_opts);

  my %text_opts = (-textvariable => \$gOpts{filetext},
                   -width        => 15);

  my $win_size =
    $top_frame->Entry(%text_opts)->pack(%pack_opts);

  return TRUE;

} # fileFrame


#############################################################################
#  normFrame
#  ---------

sub normFrame {

  my $left_frame = shift;

  my %gui_opts = (-text       => 'Normalisation options:',
                  -foreground => 'purple');

  my %pack_opts = (-padx => 2,
                   -pady => 2);

  my $norm_label =
    $left_frame->Label(%gui_opts)->pack(%pack_opts);

  # Use either an internal gene list or an external gene list file.

  %gui_opts = (-text           => 'Use external gene list',
               -foreground     => 'dark green',
               -highlightcolor => 'dark green',
               -variable       => \$gOpts{extgene},
               -anchor         => 'w');

  %pack_opts = (#-side   => 'top',
                -expand => 1,
                -anchor => 'w');

  my $gene_list_check =
    $left_frame->Checkbutton(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-text             => 'Gene list file',
               -foreground       => 'dark green',
               -activebackground => 'dark green',
               -borderwidth      => 3,
               -relief           => 'raised',
               -command          => \&openGeneList);

  my $gene_list_button =
    $left_frame->Button(%gui_opts)->pack(%pack_opts);

  # Other normalisation options.

  %gui_opts = (-text       => 'Minimum window size:',
               -foreground => 'dark green');

  my $win_label =
    $left_frame->Label(%gui_opts)->pack(%pack_opts);

  my %text_opts = (-textvariable => \$gOpts{window},
                   -width        => 15);

  my $win_size =
    $left_frame->Entry(%text_opts)->pack(%pack_opts);

  %gui_opts = (-text       => 'Maximum window size:',
               -foreground => 'dark green');

  my $maxwin_label =
    $left_frame->Label(%gui_opts)->pack(%pack_opts);

  %text_opts = (-textvariable => \$gOpts{maxwin},
                -width        => 15);

  my $maxwin_size =
    $left_frame->Entry(%text_opts)->pack(%pack_opts);

  %gui_opts = (-text       => 'Minimum normalisation block size:',
               -foreground => 'dark green');

  my $block_label =
    $left_frame->Label(%gui_opts)->pack(%pack_opts);

  %text_opts = (-textvariable => \$gOpts{blocksize},
                -width        => 15);

  my $block_size =
    $left_frame->Entry(%text_opts)->pack(%pack_opts);

  %gui_opts = (-text       => "Initial normalisation block size" .
                              "\n (accounts for absent spots):",
               -foreground => 'dark green');

  my $blockstart_label =
    $left_frame->Label(%gui_opts)->pack(%pack_opts);

  %text_opts = (-textvariable => \$gOpts{blockstart},
                -width        => 15);

  my $blockstart_size =
    $left_frame->Entry(%text_opts)->pack(%pack_opts);

  %gui_opts = (-text       => "Minimum normalisation block" .
                              "\nsize when using spikes:",
               -foreground => 'dark green');

  my $spblock_label =
    $left_frame->Label(%gui_opts)->pack(%pack_opts);

  %text_opts = (-textvariable => \$gOpts{spblocksize},
                -width        => 15);
  my $spblock_size =
    $left_frame->Entry(%text_opts)->pack(%pack_opts);

  %gui_opts = (-text       => "Initial normalisation block" .
                              "\nsize when using spikes:",
               -foreground => 'dark green');

  my $spblockst_label =
    $left_frame->Label(%gui_opts)->pack(%pack_opts);

  %text_opts = (-textvariable => \$gOpts{spblockst},
                -width        => 15);

  my $spblockst_size =
    $left_frame->Entry(%text_opts)->pack(%pack_opts);

  %gui_opts = (-text           => 'Spot normalisation file',
               -foreground     => 'dark green',
               -highlightcolor => 'dark green',
               -variable       => \$gOpts{spotnorm},
               -anchor         => 'w');

  my $spot_check =
    $left_frame->Checkbutton(%gui_opts)->pack(%pack_opts);

  #$spot_check->select();

  %gui_opts = (-text           => 'Gene normalisation file',
               -foreground     => 'dark green',
               -highlightcolor => 'dark green',
               -variable       => \$gOpts{genenorm},
               -anchor         => 'w');

  my $sig_check =
    $left_frame->Checkbutton(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-text           => 'Print all data to new file',
               -foreground     => 'dark green',
               -highlightcolor => 'dark green',
               -variable       => \$gOpts{alldata},
               -anchor         => 'w');

  my $all_data_check =
    $left_frame->Checkbutton(%gui_opts)->pack(%pack_opts);

  return TRUE;

} # normFrame


#############################################################################
#  normRoundsFrames
#  ----------------

sub normRoundsFrames {

  my $right_top_top_frame   = shift;
  my $right_top_left_frame  = shift;
  my $right_top_right_frame = shift;

  my %gui_opts = (-text       => 'Normalisation rounds:',
                  -foreground => 'purple');

  my %pack_opts = (-padx => 2,
                   -pady => 2);

  my $norm_rounds_label =
    $right_top_top_frame->Label(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-text       => "Do you wish to locally normalise\n " .
                              'all spots in the first round:',
               -foreground => 'dark green');

  my $r1norm_label =
    $right_top_left_frame->Label(%gui_opts)->pack(%pack_opts);

  %pack_opts = (-side   => 'top',
                -expand => 1,
                -anchor => 'w');

  my %r1labels = ('all-non-cont'  => "Yes (normalise all spots with\n " .
                                     'all local spots except controls)',
                  'none'          => '   No (just use spikes)');

  foreach my $but (qw(all-non-cont none) ) {

    %gui_opts = (-text           => $r1labels{$but},
                 -value          => $but,
                 -foreground     => 'dark green',
                 -highlightcolor => 'dark green',
                 -variable       => \$gOpts{round1},
                 -anchor         => 'w');

    my $r1norm_but =
      $right_top_left_frame->Radiobutton(%gui_opts)->pack(%pack_opts);

  }

  %gui_opts = (-text       => "Type of normalisation with spikes:",
               -foreground => 'dark green');

  my $r2norm_label =
    $right_top_right_frame->Label(%gui_opts)->pack(%pack_opts);

  my %r2labels = ('none'   => "    Don't correct with spikes",
                  'local'  => "Locally normalise/correct\n " .
                              'all spots with corrected spikes',
                  'global' => "Globally normalise/correct\n " .
                              'all spots with corrected spikes');

  foreach my $but (qw(none local global) ) {

    %gui_opts = (-text           => $r2labels{$but},
                 -value          => $but,
                 -foreground     => 'dark green',
                 -highlightcolor => 'dark green',
                 -variable       => \$gOpts{round2},
                 -anchor         => 'w');

    my $r2norm_but =
      $right_top_right_frame->Radiobutton(%gui_opts)->pack(%pack_opts);

  }

  return TRUE;

} # normRoundsFrames


#############################################################################
#  filterFrame
#  -----------

sub filterFrame {

  my $right_bottom_frame = shift;

  my %gui_opts = (-text       => 'Filtering options:',
                  -foreground => 'purple');

  my %pack_opts = (-padx => 2,
                   -pady => 2);

  my $filt_label =
    $right_bottom_frame->Label(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-text           => 'Background SD cut-offs',
               -foreground     => 'dark green',
               -highlightcolor => 'dark green',
               -variable       => \$gOpts{bgcut},
               -anchor         => 'w');

  %pack_opts = (-side   => 'top',
                -expand => 1,
                -anchor => 'w');

  my $bg_cut_offs =
    $right_bottom_frame->Checkbutton(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-text       => 'Lower cut-off:',
               -foreground => 'dark green');

  my $lower_label =
    $right_bottom_frame->Label(%gui_opts)->pack(%pack_opts);

  my %text_opts = (-textvariable => \$gOpts{lowerbg},
                   -width        => 15);

  my $lower_bg =
    $right_bottom_frame->Entry(%text_opts)->pack(%pack_opts);

  %gui_opts = (-text       => 'Upper cut-off:',
               -foreground => 'dark green');

  my $upper_label =
    $right_bottom_frame->Label(%gui_opts)->pack(%pack_opts);

  %text_opts = (-textvariable => \$gOpts{upperbg},
                -width        => 15);

  my $upper_bg = $right_bottom_frame->Entry(%text_opts)->pack(%pack_opts);

  %gui_opts = (-text       => "Diameter cut-off\n  " .
                              '(0 means no cut-off used):',
               -foreground => 'dark green');

  my $diam_label =
    $right_bottom_frame->Label(%gui_opts)->pack(%pack_opts);

  %text_opts = (-textvariable => \$gOpts{diamcut},
                -width        => 15);

  my $diam_cut =
    $right_bottom_frame->Entry(%text_opts)->pack(%pack_opts);

  return TRUE;

} # filterFrame


#############################################################################
#  doGui
#  -----

sub doGui {

  # Overall headers.

  $gTk->title('Local Normalisation');

  my %gui_opts = (-text        => 'Local Normalisation for GenePix Results',
                  -foreground  => 'red',
                  -height      => 3,
                  -width       => 3,
                  -borderwidth => 3,
                  -relief      => 'groove');

  my %pack_opts = (-fill => 'both');

  my $header_label =
    $gTk->Label(%gui_opts)->pack(%pack_opts);

  # Create frames to put widgets into.

  %gui_opts = (-height      => 3,
               -width       => 3,
               -borderwidth => 3,
               -relief      => 'groove');

  %pack_opts = (-padx => 2,
                -pady => 2,
                -side => 'top');

  my $top_frame =
    $gTk->Frame(%gui_opts)->pack(%pack_opts);

  %pack_opts = (-padx => 2,
                -pady => 2,
                -side => 'bottom');

  my $bottom_frame =
    $gTk->Frame(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-height      => 20,
               -width       => 20,
               -borderwidth => 3,
               -relief      => 'groove');

  %pack_opts = (-ipadx => 3,
                -fill  => 'both');

  my $middle_frame =
    $gTk->Frame(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-height      => 10,
               -width       => 20,
               -borderwidth => 3,
               -relief      => 'groove');

  %pack_opts = (-padx => 2,
                -pady => 2,
                -side => 'left');

  my $left_frame =
    $middle_frame->Frame(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-height      => 20,
               -width       => 20,
               -borderwidth => 0);

  %pack_opts = (-padx => 2,
                -pady => 2,
                -side => 'right');

  my $right_frame =
    $middle_frame->Frame(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-height      => 20,
               -width       => 20,
               -borderwidth => 3,
               -relief      => 'groove');

  %pack_opts = (-padx => 2,
                -pady => 2,
                -side => 'top');

  my $right_top_frame =
    $right_frame->Frame(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-height => 10,
               -width  => 20);

  %pack_opts = (-padx => 2,
                -pady => 2,
                -side => 'top');

  my $right_top_top_frame =
    $right_top_frame->Frame(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-height      => 10,
               -width       => 20,
               -borderwidth => 3,
               -relief      => 'groove');

  %pack_opts = (-padx => 2,
                -pady => 2,
                -side => 'left');

  my $right_top_left_frame =
    $right_top_frame->Frame(%gui_opts)->pack(%pack_opts);

  %gui_opts = (-height      => 10,
               -width       => 20,
               -borderwidth => 3,
               -relief      => 'groove');

  %pack_opts = (-padx => 2,
                -pady => 2,
                -side => 'right');

  my $right_top_right_frame =
    $right_top_frame->Frame(%gui_opts)->pack(%pack_opts);

  %pack_opts = (-padx => 2,
                -pady => 2,
                -side => 'bottom');

  my $right_bottom_frame =
    $right_frame->Frame(%gui_opts)->pack(%pack_opts);

  # Input file(s).

  fileFrame($top_frame)
    or Die('Cannot create input file(s) frame.');

  # Normalisation options.

  normFrame($left_frame)
    or Die('Cannot create normalisation options frame.');

  # Options for how many rounds of normalisations occur with which spots.

  normRoundsFrames($right_top_top_frame,
                   $right_top_left_frame,
                   $right_top_right_frame)
    or Die('Cannot create normalisation rounds frames.');

  # Filtering options.

  filterFrame($right_bottom_frame)
    or Die('Cannot create filtering options frame.');

  # Start button.

  %gui_opts = (-text             => 'Start analysis',
               -foreground       => 'red',
               -activebackground => 'red',
               -borderwidth      => 3,
               -relief           => 'raised',
               -cursor           => 'mouse',
               -command          => \&startAnal);

  my $start_button =
    $bottom_frame->Button(%gui_opts)->pack();

  %gui_opts = (-text    => 'Exit',
               -command => sub { exit; },
               -cursor  => 'man');

  %pack_opts = (-side => 'bottom',
                -fill => 'x');

  # Exit button.

  my $exit_button =
    $bottom_frame->Button(%gui_opts)->pack(%pack_opts);

  # Tk MainLoop command.

  MainLoop;

  return TRUE;

} # doGui


#############################################################################
#  MAIN:
#  -----

MAIN: {

  undef %gOpts;

  setGlobals()
    or Die('Cannot set global variables.');

  # Set a few things depending on whether a GUI version
  #   or a command-line version is being used.

  if (@ARGV == 0) {

    $gTk = new MainWindow;

    $gOpts{tk} = TRUE;

  } else {

    $gTk = undef;

    $gOpts{tk} = FALSE;

  }

  # UNIX is set as the default OS,
  #   unless $ENV{OS} indicates it is Windows.

  if (exists $ENV{OS} and defined $ENV{OS}) {

    if ($ENV{OS} =~ /^Windows/i) {

      $gOpts{os} = WINDOWS;

    } else {

      $gOpts{os} = UNIX;

    }

  } else {

    $gOpts{os} = UNIX;

  }

  # Set default options before user interface starts.

  setOpts()
    or Die('Cannot set default options.');

  if ($gOpts{tk} == TRUE) {

    # Tk GUI version.

    doGui()
      or exit 8;

  } else {

    # Command line version.

    doCmd()
      or exit 8;

  }

  exit 0;

} # MAIN:
